import json
import pandas
import numpy as np


def peaks_from_fasta(filename):
    with open(filename) as f:
        parts = (line.split(maxsplit=1) for line in f if line.startswith(">"))
        entries = [json.loads(part[1]) for part in parts]
        return pandas.DataFrame(entries)


def df_from_fimo(filename):
    return pandas.read_csv(filename, sep="\t").set_index("sequence_name")

def df_from_subgraphs(filename):
    count_edges = lambda x: np.count_nonzero(x.data)
    count_nodes = lambda x: x.shape[0]-1
    subgraphs = {name: graph[()] for name, graph in np.load(filename).items()}
    return pandas.DataFrame([{"name": name, "edges": count_edges(graph), "nodes": count_nodes(graph)}
                             for name, graph in subgraphs.items()]).set_index("name")
    
def df_from_setsummary(filename):
    names, parts = zip(*(line.split(", (", 1) for line in open(filename) if not line.startswith("#")))
    # names = ["peak"+name for name in names]
    parts = [part.split(", ", 2) for part in parts]
    ratios = [(float(a), float(b)) for a, b, _ in parts]
    df = pandas.DataFrame([{"coverage": r[0], "reads": r[1]}
                           for r in ratios])
    return df

def combine_dfs(folder, name):
    peaks = peaks_from_fasta(folder + name + "_sequences_summits_unique.fasta")
    fimo = df_from_fimo(folder + "fimo_graph_chr"+ name + "/fimo.txt")
    tmp = df_from_subgraphs(folder+name+"_sub_graphs.graphs.npz")
    coverage = df_from_setsummary(folder + name + "_sequences_summits_unique.setsummary")
    combined = pandas.concat([peaks, coverage], axis=1, join="inner")
    combined = combined.set_index("unique_id")
    subgraphs = pandas.concat([combined, tmp], axis=1, join="inner")
    match_ids = set(fimo.index.unique())
    subgraphs["MATCH"] = [name in match_ids for name in subgraphs.index.values]
    print(subgraphs.shape)
    return subgraphs
    matches = subgraphs.loc[subgraphs["MATCH"]]
    nonmatches = subgraphs.loc[~subgraphs["MATCH"]]
    print("MATCHES", (matches["edges"]-matches["nodes"]).mean(), matches.shape[0])
    print("NONMAtCHES", (nonmatches["edges"]-nonmatches["nodes"]).mean(), nonmatches.shape[0])


def motif_haplotype_analysis(folder, name):
    coverage = df_from_setsummary(folder + name + "_sequences_summits_unique.setsummary")


if __name__ == "__main__":
    dfs = [combine_dfs("./", str(c)) for c in range(1, 6)]
    combined = pandas.concat(dfs, axis=0, keys=list(range(1, 6)))
    print(combined["MATCH"].mean())
    print(combined["MATCH"].sum())
    tmp = combined.loc[(combined["coverage"]/combined["reads"])>=0.8]
    print(tmp)
    print(tmp["MATCH"].mean())
    print(combined.loc[(combined["coverage"]/combined["reads"])<0.8]["MATCH"].mean())
    # for c in (str(i) for i in range(1,6)):
    #     combine_dfs("./", # "/home/ivar/dev/graph_peak_caller/benchmarking/data/ARABIDOPSIS_ERF115_SRR931836/1/",
    #                 c)

    # peaks = peaks_from_fasta("/home/ivar/dev/graph_peak_caller/benchmarking/data/ARABIDOPSIS_ERF115_SRR931836/1/unique_graph.fasta")
    # fimo = df_from_fimo("/home/ivar/dev/graph_peak_caller/benchmarking/data/ARABIDOPSIS_ERF115_SRR931836/1/fimo_unique_graph/fimo.txt")
    # subgraphs = df_from_subgraphs("/home/ivar/dev/graph_peak_caller/benchmarking/data/ARABIDOPSIS_ERF115_SRR931836/1/1_sub_graphs.graphs.npz")
    # print(subgraphs)




import numpy as np
from .sparse_pileup import SparsePileup
from offsetbasedgraph import Position


class ShiftParameters(object):
    def __init__(self, mfold, winsize=10):
        self.mfold = mfold
        self.winsize = winsize


class PeakModel(object):
    def __init__(self, graph, intervals, shift_parameters):
        self.graph = graph
        self.intervals = intervals
        self.parameters = shift_parameters
        self.pileups = {}

    def _create_pileups(self):
        for direction in [+1, -1]:
            self.pileups[direction] = SparsePileup.from_intervals(
                (i for i in self.intervals if i.direction == direction))

    def _find_peaks(self):
        for direction in [+1, -1]:
            self.peaks[direction] = self.pileups[direction].threshold(
                self.parameters.mfold)

    def _find_centers(self):
        intervals = {direction: peaks.find_valued_intevals(True)
                     for direction, peaks in self.peaks.items()}

        self.centers = {
            direction: [self.pileups[direction].find_max(interval)
                        for interval in intervals]
            for direction, intervals in intervals.items()}

    def _rec_find_pairs_on_edge(self):
        for node_id, centers_array in self.centers_arrays[node]:
            last_center = centers_array[-1]
            if last_center % 3 != 0:
                continue
            last_center //= 3
            if self.graph.node_size(node_id) - last_center > self.parameters.win_size:
                pass

    def _find_pairs_on_edge(self):
        pairs = []
        for node, edge_list in self.graph.adj_list.items():
            pos_center = self.centers_arrays[node][-1]
            if pos_center % 3 != 0:
                continue
            pos_center //= 3
            distance_to_end = self.graph.node_size(node) - pos_center
            if distance_to_end > self.parameters.win_size:
                continue
            for next_node in edge_list:
                neg_center = self.centers_arrays[node]
                if start % 3 != 1:
                    continue
                neg_center //= 3
                if neg_center + distance_to_end <= self.parameters.win_size:
                    pairs.append(Position(node, pos_center),
                                 Position(next_node, neg_center))
        return pairs

    def _find_paired_peaks_on_node(self, centers_array):
        centers_array.sort()
        diffs = np.diff(centers_array)
        pair_idxs = np.where(diffs % 3 == 1 and diffs < (self.params.win_size * 3))[0]
        return centers_array[pair_idxs], centers_array[pair_idxs+1]

    def _find_paired_peaks(self):
        centers_dicts = {}
        for direction in [+1, -1]:
            centers_dicts[direction] = defaultdict(list)
            for center in self.centers[direction]:
                centers_dicts[direction][center.region_path].append(
                    center.offset)

        centers_arrays = {node_id: np.array(self.centers_dicts[+1][node_id] +
                                            self.centers_dicts[-1][node_id])
                          for node_id in set(centers_dicts[+1].keys() + centers_dicts[-1].keys())}

        for node_id, center_array in centers_arrays.items():
            centers_arrays *= 3
            centers_arrays[len(self.centers_dicts[-1]):] += 1

        pairs = []
        for node_id, array in centers_arrays.items():
            pos_centers, neg_centers = self._find_paired_peaks(array)
            pairs.extend([
                (Position(node_id, pos_center),
                 Position(node_id, neg_center))
                for pos_center, neg_center in
                zip(pos_centers, neg_centers)])

        pairs.extend(self._find_pairs_on_edge())

    def _get_windows(self):
        self.



import logging
import sys
import warnings
import numpy as np
import scipy

def customwarn(message, category, filename, lineno, file=None, line=None):
    sys.stdout.write(warnings.formatwarning(message, category, filename, lineno))

def set_logging_config(chosen_level=1):

    # Remove handlers already set (if not, new config will not be set)
    root = logging.getLogger()
    if root.handlers:
        for handler in root.handlers:
            root.removeHandler(handler)

    class InfoFilter(logging.Filter):
        def filter(self, rec):
            return rec.levelno in (logging.WARNING, logging.DEBUG, logging.INFO)

    logger = logging.getLogger(__name__)
    logger.setLevel(logging.DEBUG)

    h1 = logging.StreamHandler(sys.stdout)
    h1.setLevel(logging.DEBUG)
    h1.addFilter(InfoFilter())
    formatter = logging.Formatter('%(asctime)s, %(levelname)s: %(message)s')
    h1.setFormatter(formatter)
    h2 = logging.StreamHandler()
    h2.setLevel(logging.WARNING)
    h2.setFormatter(formatter)

    logger.addHandler(h1)
    logger.addHandler(h2)
    logger.propagate = False

    levels = {0: logging.WARNING,
              1: logging.INFO,
              2: logging.DEBUG}

    if chosen_level not in levels:
        print("Error. Invalid verbosity level %s. Must be 0, 1 or 2" % (chosen_level))
        raise ValueError("Invalid logging level set by -v/--verbose. Choose between 0, 1 and 2.")

    use_level = levels[chosen_level]
    logging.basicConfig(
        level=use_level,
        format="%(asctime)s, %(levelname)s: %(message)s",
        handlers=[h1, h2]
    )

    warnings.showwarning = customwarn
    np.seterr(all='print')
    scipy.seterr(all='print')



import json
import offsetbasedgraph as obg
import logging
from collections import defaultdict
import numpy as np
from .analysis.nongraphpeaks import NonGraphPeakCollection, NonGraphPeak
from .summits import find_summits


class Peak(obg.DirectedInterval):
    def __init__(self, start_position, end_position,
                 region_paths=None, graph=None, direction=None, score=0, unique_id=None):
        super().__init__(start_position, end_position,
                         region_paths, graph, direction)
        self.score = score
        self.unique_id = unique_id
        self.info = (0, 0)

    def __str__(self):
        return super().__str__() + " [%s]" % self.score

    @classmethod
    def from_interval_and_score(cls, interval, score):
        return cls(interval.start_position, interval.end_position,
                   interval.region_paths, interval.graph, interval.direction,
                   score)

    def set_score(self, score):
        assert isinstance(score, float), "Score %s is invalid, type %s" % (score, type(score))
        self.score = score

    def to_file_line(self):
        object = {"start": int(self.start_position.offset),
                  "end": int(self.end_position.offset),
                  "region_paths": [int(r) for r in self.region_paths],
                  "direction": int(self.direction),
                  "average_q_value": float(self.score),
                  "unique_id": str(self.unique_id),
                  "is_ambigous": int(self.info[1]),
                  "is_diff": int(self.info[0])
                  }
        try:
            d = json.dumps(object)
        except:
            for k, v in object.items():
                print(k, v, type(v))
            raise
        return d

    @classmethod
    def from_file_line(cls, line, graph=None):
        object = json.loads(line)
        unique_id = None
        if "unique_id" in object:
            unique_id = object["unique_id"]
        obj = cls(object["start"], object["end"], object["region_paths"],
                  direction=object["direction"], graph=graph,
                  score=object["average_q_value"],
                  unique_id=unique_id)
        obj.info = (object["is_diff"], object["is_ambigous"])
        return obj

    def to_approx_linear_peak(self, linear_path, chromosome):
        # Assuming graph is partially ordered DAG
        intersecting_nodes = set(self.region_paths).intersection(linear_path.nodes_in_interval())
        intersecting_nodes = sorted(list(intersecting_nodes))

        first_node = intersecting_nodes[0]
        start_offset = 0
        if first_node == self.region_paths[0]:
            start_offset = self.start_position.offset

        last_node = intersecting_nodes[-1]
        end_offset = 0
        if last_node == self.region_paths[-1]:
            end_offset = self.end_position.offset

        linear_start_pos = linear_path.get_offset_at_node(first_node) + start_offset
        linear_end_pos = linear_path.get_offset_at_node(last_node) + end_offset
        return NonGraphPeak(chromosome, linear_start_pos, linear_end_pos, score=self.score)

    def get_subinterval(self, start_offset, end_offset):
        subinterval = obg.Interval.get_subinterval(self, start_offset, end_offset)
        peak = Peak(subinterval.start_position, subinterval.end_position,
                    subinterval.region_paths, graph=self.graph, score=self.score,
                    unique_id=self.unique_id)
        peak.info = self.info
        return peak


class PeakCollection(obg.IntervalCollection):
    interval_class = Peak

    def cut_around_summit(self, q_values, n_base_pairs_around=60):
        def get_summit(peak):
            peak_qvalues = q_values.get_interval_values(peak)
            max_positions = np.flatnonzero(peak_qvalues == np.max(peak_qvalues))
            summit_position = np.partition(max_positions, max_positions.size//2)[max_positions.size//2]
            return peak.get_subinterval(
                max(0, summit_position - n_base_pairs_around),
                min(summit_position + n_base_pairs_around, peak.length()))

        def get_summit_fancy(peak):
            peak_qvalues = q_values.get_interval_values(peak)
            summits = find_summits(peak_qvalues)
            if not summits.size:
                return get_summit(peak)
            summit_values = peak_qvalues[summits]
            summit_position = summits[np.argmax(summit_values)]
            return peak.get_subinterval(
                max(0, summit_position - n_base_pairs_around),
                min(summit_position + n_base_pairs_around, peak.length()))

        self.intervals = [get_summit(peak) for peak in self.intervals]
        for peak in self.intervals:
            assert peak.length() <= n_base_pairs_around * 2

    @classmethod
    def _is_in_graph(cls, peak, chrom, start_offset, end_offset):
        if peak.chromosome != chrom:
            return False
        if (peak.start < start_offset or peak.end > end_offset):
            return False
        return True

    def create_node_index(self):
        # Creates an index from region path to intervals touching the rp,
        # making it fast to do approx. overlap
        index = defaultdict(list)
        for peak in sorted(self.intervals, key=lambda x: x.score, reverse=True):
            for rp in peak.region_paths:
                index[abs(rp)].append(peak)

        self._index = index

    def get_all_overlapping(self, interval):
        return {peak for rp in interval.region_paths for peak in self._index[abs(rp)]}

    def approx_contains_part_of_interval(self, interval, visited=None):
        assert hasattr(self, "_index"), "Create index first by calling create_node_index()"
        visited = visited if visited is not None else set([])
        for rp in interval.region_paths:
            touching = [peak for peak in self._index[abs(rp)]
                        if peak.unique_id not in visited]
            if len(touching):
                return list(touching)

        return False

    def which_approx_contains_part_of_interval(self, interval):
        assert hasattr(self, "_index"), "Create index first by calling create_node_index()"
        for rp in interval.region_paths:
            containing = self._index[abs(rp)]
            if len(containing) > 0:
                return containing[0]

    @classmethod
    def create_from_nongraph_peak_collection(
            cls, ob_graph, peak_collection,
            linear_path_interval,
            graph_region=None):
        peaks = peak_collection.peaks
        intervals_on_graph = []
        i = 0
        graph_start_offset = 0 if graph_region is None else graph_region.start
        for peak in peaks:
            start = peak.start - graph_start_offset
            end = peak.end - graph_start_offset
            end = min(end, linear_path_interval.length())
            if graph_region is not None:
                if not cls._is_in_graph(peak, graph_region.chromosome,
                                        graph_region.start,
                                        graph_region.end):
                    logging.info("Filtered out peak")
                    continue
            if i % 10000 == 0:
                logging.info("%d peaks processed" % (i))
            i += 1
            linear_interval = linear_path_interval.get_exact_subinterval(start, end)
            linear_interval.graph = ob_graph
            graph_peak = Peak.from_interval_and_score(
                linear_interval, peak.score)

            assert graph_peak.length() == end - start, "Graph peak length %d != linear peak length %d for peak %s" % (graph_peak.length(), end-start, graph_peak)
            assert graph_peak.end_position.offset <= ob_graph.blocks[graph_peak.end_position.region_path_id].length()

            graph_peak.unique_id = peak.unique_id
            graph_peak.sequence = peak.sequence
            intervals_on_graph.append(graph_peak)

        return cls(intervals_on_graph)

    def contains_interval(self, interval):
        for i in self.intervals:
            if i == interval:
                return True

        return False

    def get_similar_intervals(self, interval, allowed_mismatches):
        similar = []
        for i in self.intervals:
            if i.is_approx_equal(interval, allowed_mismatches):
                similar.append(i)

        return similar

    def get_identical_intervals(self, other_peak_collection):
        identical_intervals = []
        for interval in self.intervals:
            if other_peak_collection.contains_interval(interval):
                identical_intervals.append(interval)

        return identical_intervals

    def get_overlapping_intervals(self, interval, minimum_overlap=1):
        overlapping = []
        for i in self.intervals:
            if i.overlaps(interval, minimum_overlap=minimum_overlap):
                overlapping.append(i)
        return overlapping

    def to_approx_linear_peaks(self, linear_path, chromosome):
        linear_peaks = []
        i = 0
        for peak in self.intervals:
            if i % 500 == 0:
                logging.info("Converting peak %d" % i)
            i += 1
            linear_peaks.append(peak.to_approx_linear_peak(linear_path, chromosome))

        return NonGraphPeakCollection(linear_peaks)

    def to_fasta_file(self, file_name, sequence_graph):
        from .peakfasta import PeakFasta
        PeakFasta(sequence_graph).save_intervals(file_name, self)

    @classmethod
    def from_fasta_file(cls, file_name, graph=None):
        f = open(file_name)
        peaks = []
        while True:
            header = f.readline()
            sequence = f.readline()
            if not sequence:
                break

            header = header.split(maxsplit=1)
            id = header[0].replace(">", "")
            interval_json = header[1]
            peak = Peak.from_file_line(interval_json)
            peak.unique_id = id
            peak.sequence = sequence.strip()
            peak.graph = graph
            peaks.append(peak)
        if graph is not None:
            avg_peak_size = np.mean([p.length() for p in peaks])
            logging.info("Avg peak size: %2.f" % avg_peak_size)

        f.close()

        return cls(peaks)




import numpy as np
import logging
import offsetbasedgraph as obg
from pyvg.conversion import vg_json_file_to_interval_collection
from . import CallPeaks
from .sparsepvalues import PToQValuesMapper
from .sparsediffs import SparseValues
from .intervals import Intervals, UniqueIntervals

from .peakfasta import PeakFasta


class MultipleGraphsCallpeaks:

    def __init__(self, graph_names, graph_file_names,
                 samples,
                 controls, linear_maps,
                 config, reporter,
                 sequence_retrievers=None,
                 stop_after_p_values=False,
                 ):
        self._config = config
        self._reporter = reporter
        self.names = graph_names
        self.graph_file_names = graph_file_names
        self.linear_maps = linear_maps
        self.sequence_retrievers = sequence_retrievers
        self.samples = samples
        self.controls = controls
        self.stop_after_p_values = stop_after_p_values
        if self.stop_after_p_values:
            logging.info("Will only run until p-values have been computed.")

    @classmethod
    def count_number_of_unique_reads(cls, sample_reads):
        n_unique = 0
        for reads in sample_reads:
            logging.info("Processing sample")

            interval_hashes = set()
            n_duplicates = 0
            i = 0
            for interval in reads.intervals:
                if i % 20000 == 0:
                    logging.info("%d reads processed" % i)
                i += 1
                hash = interval.hash(ignore_end_pos=True)
                if hash in interval_hashes:
                    n_duplicates += 1
                else:
                    n_unique += 1
                interval_hashes.add(hash)

            logging.info("Found %d duplicates" % n_duplicates)

        logging.info("In total %d unique reads" % n_unique)
        return n_unique

    @classmethod
    def from_file_base_names(cls):
        # Create lists of names, graphs, sample, control linear maps
        pass

    def run(self):
        self.run_to_p_values()
        if self.stop_after_p_values:
            logging.info("Stopping, as planned, after p-values")
            return
        self.create_joined_q_value_mapping()
        self.run_from_p_values()

    def get_intervals(self, sample, control, graph):
        if isinstance(sample, Intervals) or isinstance(sample, UniqueIntervals):
            logging.info("Sample is already intervalcollection.")
            return sample, control
        elif sample.endswith(".intervalcollection"):
            try:
                sample = obg.IntervalCollection.from_file(
                    sample, graph=graph)
            except OSError:
                sample = obg.IntervalCollection.from_file(
                    sample, graph=graph, text_file=True)
            try:
                control = obg.IntervalCollection.from_file(
                    control, graph=graph)
            except:
                control = obg.IntervalCollection.from_file(
                    control, graph=graph, text_file=True)
        else:
            logging.info("Creating interval collections from files")
            sample = vg_json_file_to_interval_collection(sample, graph)
            control = vg_json_file_to_interval_collection(control, graph)

        if self._config.keep_duplicates:
            logging.warning("Keeping duplicates. Should only be used for testing.")
            return Intervals(sample), Intervals(control)
        else:
            return UniqueIntervals(sample), UniqueIntervals(control)

    def run_to_p_values(self):
        for name, graph_file_name, sample, control, lin_map in \
            zip(self.names, self.graph_file_names,
                self.samples, self.controls, self.linear_maps):
            logging.info("Running to p values, %s" % name)
            ob_graph = obg.Graph.from_file(
                graph_file_name)
            sample, control = self.get_intervals(sample, control, ob_graph)
            config = self._config.copy()
            config.linear_map_name = lin_map
            caller = CallPeaks(ob_graph, config,
                               self._reporter.get_sub_reporter(name))
            caller.run_to_p_values(sample, control)
            logging.info("Done until p values.")
            logging.info("In total %d duplicates were removed from sample" % sample.n_duplicates)

    def create_joined_q_value_mapping(self):
        mapper = PToQValuesMapper.from_files(self._reporter._base_name)
        self._q_value_mapping = mapper.get_p_to_q_values()

    def run_from_p_values(self, only_chromosome=None):
        for i, name in enumerate(self.names):
            logging.info("Name: %s" % name)
            if only_chromosome is not None:
                if only_chromosome != name:
                    logging.info("Skipping %s" % str(name))
                    continue
            graph_file_name = self.graph_file_names[i]
            ob_graph = obg.Graph.from_numpy_file(
                graph_file_name)
            assert ob_graph is not None
            caller = CallPeaks(ob_graph, self._config,
                               self._reporter.get_sub_reporter(name))
            caller.p_to_q_values_mapping = self._q_value_mapping
            if name != "":
                name += "_"
            caller.p_values_pileup = SparseValues.from_sparse_files(
                self._reporter._base_name + name + "pvalues")
            caller.touched_nodes = set(np.load(
                self._reporter._base_name + name + "touched_nodes.npy"))
            caller.get_q_values()
            caller.call_peaks_from_q_values()
            if self.sequence_retrievers is not None:
                try:
                    sequencegraph = self.sequence_retrievers.__next__()
                except FileNotFoundError:
                    logging.warning("Could not find sequence graphs. Will not store max paths.")
                    continue

                PeakFasta(sequencegraph).write_max_path_sequences(
                  self._reporter._base_name + name + "sequences.fasta", caller.max_path_peaks)
                #caller.save_max_path_sequences_to_fasta_file(
                #    "sequences.fasta", self.sequence_retrievers.__next__())



import numpy as np
import logging
from .linearintervals import LinearIntervalCollection
from ..sparsediffs import SparseDiffs


def get_reference_map(graph, reference_path):
    linear_map = LinearMap.from_graph(graph)
    ref_starts = linear_map._node_starts[reference_path]
    ref_ends = linear_map._node_ends[reference_path]
    def mapping_func(intervals):
        linear_intervals = LinearMap.map_interval_collection(intervals)
        start_nodes = np.searchsorted(ref_starts, linear_intervals.starts)
        start_offsets = linear_intervals.starts-LinearMap._node_starts[start_nodes]
        
        ends = np.searchsorted(ref_ends, linear_intervals.ends)
    


class LinearMap:
    def __init__(self, node_starts, node_ends, graph):
        self._node_starts = np.asanyarray(node_starts)
        self._node_ends = np.asanyarray(node_ends)
        assert np.all(self._node_starts < self._node_ends),\
            np.where(self._node_starts >= self._node_ends)
        self._length = self._node_ends[-1]
        self._graph = graph

    def __eq__(self, other):
        if not np.all(self._node_starts == other._node_starts):
            return False
        return np.all(self._node_ends == other._node_ends)

    def __repr__(self):
        return "%s:%s" % (self._node_starts, self._node_ends)

    def get_node_start(self, node_id):
        return self._node_starts[node_id-self._graph.min_node]

    def get_node_end(self, node_id):
        return self._node_ends[node_id-self._graph.min_node]

    def get_scale_and_offset(self, node_id):
        linear_length = self.get_node_end(node_id) \
                        - self.get_node_start(node_id)
        node_length = self._graph.node_size(node_id)
        scale = linear_length/node_length
        offset = self.get_node_start(node_id)
        return scale, offset

    def map_interval_collection(self, interval_collection):
        starts = []
        ends = []
        for interval in interval_collection:
            start, end = self.map_graph_interval(interval)
            starts.append(start)
            ends.append(end)
        return LinearIntervalCollection(starts, ends)

    def map_graph_interval(self, interval):
        start_pos = self.graph_position_to_linear(interval.start_position)
        end_pos = self.graph_position_to_linear(interval.end_position)
        return start_pos, end_pos

    def graph_position_to_linear(self, position):
        node_id = abs(position.region_path_id)-self._graph.min_node
        node_start = self._node_starts[node_id]
        node_end = self._node_ends[node_id]
        node_size = self._graph.node_size(position.region_path_id)
        scale = (node_end-node_start) / node_size
        if position.region_path_id > 0:
            return node_start + scale*position.offset
        else:
            return node_end - scale*position.offset

    def to_sparse_pileup(self, unmapped_indices_dict, min_value=0):
        all_indices = []
        all_values = []
        i = 0
        node_idxs = self._graph.node_indexes
        min_idx = self._graph.min_node
        for i in range(node_idxs.size-1):
            if i % 100000 == 0:
                logging.info("Processing node %d" % i)
            node_id = i+min_idx
            if node_id not in unmapped_indices_dict:
                all_indices.append(node_idxs[i])
                all_values.append(min_value)
                continue
            unmapped_indices = unmapped_indices_dict[node_id]
            scale, offset = self.get_scale_and_offset(node_id)
            new_idxs = [(idx-offset)//scale+node_idxs[i]
                        for idx in unmapped_indices.indices]
            new_idxs[0] = max(node_idxs[i], new_idxs[0])
            all_indices.extend(new_idxs)
            all_values.extend(unmapped_indices.values)
        return SparseDiffs(
            np.array(all_indices, dtype="int"),
            np.diff(np.r_[0, all_values]))

    @classmethod
    def from_graph(cls, graph):
        logging.info("Getting topologically sorted nodes")
        node_ids = list(graph.get_topological_sorted_node_ids())
        logging.info("Finding starts and ends")
        starts = cls.find_starts(graph, node_ids)
        ends = cls.find_ends(graph, node_ids[::-1])
        return cls(starts, ends, graph)

    @classmethod
    def from_file(cls, filename, graph):
        obj = np.load(filename)
        return cls(obj["starts"], obj["ends"], graph)

    def to_file(self, filename):
        np.savez(filename, starts=self._node_starts,
                 ends=self._node_ends)

    @staticmethod
    def find_starts(graph, node_ids=None):
        if node_ids is None:
            node_ids = list(graph.get_topological_sorted_node_ids())
        max_dists = np.zeros(len(node_ids))
        n_processed = 0
        for node_id in node_ids:
            if n_processed % 500000 == 0:
                logging.info("%d nodes processed" % n_processed)
            n_processed += 1
            assert graph.node_size(node_id) > 0
            i = node_id - graph.min_node
            cur_dist = max_dists[i] + graph.node_size(node_id)
            for next_node in graph.adj_list[node_id]:
                j = next_node-graph.min_node
                max_dists[j] = max(cur_dist, max_dists[j])
        return max_dists

    @staticmethod
    def find_ends(graph, node_ids=None):
        adj_list = graph.reverse_adj_list
        if node_ids is None:
            node_ids = list(graph.get_sorted_node_ids(reverse=True))
        max_dists = np.zeros(len(node_ids))
        n_processed = 0
        for node_id in node_ids:
            if n_processed % 500000 == 0:
                logging.info("%d nodes processed" % n_processed)
            n_processed += 1
            i = node_id - graph.min_node
            cur_dist = max_dists[i] + graph.node_size(node_id)
            for next_node in adj_list[-node_id]:
                j = abs(next_node)-graph.min_node
                max_dists[j] = max(cur_dist, max_dists[j])
        linear_length = max_dists[0] + graph.node_size(graph.min_node)
        return linear_length - max_dists

if __name__ == "__main__":
    import offsetbasedgraph as obg
    graph = obg.GraphWithReversals.from_numpy_file("graph.nobg")
    LinearMap.from_graph(graph)



import logging
import os
import numpy as np

import offsetbasedgraph as obg
from pyvg.conversion import vg_json_file_to_interval_collection

from . import Configuration, CallPeaks
from .multiplegraphscallpeaks import MultipleGraphsCallpeaks
from .util import create_linear_map
from .peakfasta import PeakFasta
from .reporter import Reporter
from .intervals import UniqueIntervals
from .shiftestimation import MultiGraphShiftEstimator
import sys


def estimate_read_length(file_name, graph_name):
    graph = obg.Graph.from_file(graph_name)
    if file_name.endswith(".intervalcollection"):
        intervals = obg.IntervalCollection.create_generator_from_file(
            file_name, graph=graph)
    else:
        intervals = vg_json_file_to_interval_collection(file_name, graph)
    return int(np.median([i.length() for i in intervals]))


def get_confiugration(args):
    config = Configuration()
    config.has_control = args.control is not None
    if args.linear_map is None:
        config.linear_map_name = args.graph_file_name.split(".")[0] + "_linear_map.npz"
    else:
        config.linear_map_name = args.linear_map
    config.fragment_length = int(args.fragment_length)
    config.read_length = int(args.read_length)
    if args.q_value_threshold is not None:
        config.q_value_threshold = float(args.q_value_threshold)
    return config


def get_intervals(args):
    iclass = UniqueIntervals  # Use Intervals to skip filter dup
    samples = iclass(parse_input_file(args.sample, args.graph))
    control_name = args.sample if args.control is None else args.control
    controls = iclass(parse_input_file(control_name, args.graph))
    return samples, controls


def get_callpeaks(args):
    config = get_confiugration(args)
    find_or_create_linear_map(args.graph, config.linear_map_name)
    out_name = args.out_name if args.out_name is not None else ""
    reporter = Reporter(out_name)
    return CallPeaks(args.graph, config, reporter)


def parse_input_file(input, graph):
    logging.info("Parsing input file %s" % input)
    try:
        if isinstance(input, obg.IntervalCollection):
            return input
        elif input.endswith(".json"):
            intervals = vg_json_file_to_interval_collection(input, graph)
            return intervals
        else:
            try:
                intervals = obg.IntervalCollection.from_file(
                    input, graph=graph)
            except OSError:
                intervals = obg.IntervalCollection.from_file(
                    input, graph=graph, text_file=True)
            return intervals
    except FileNotFoundError as e:
        logging.debug(e)
        logging.critical("Input file %s not found. Aborting. " % input)
        sys.exit(1)


def run_callpeaks_interface(args):
    caller = get_callpeaks(args)
    inputs, controls = get_intervals(args)
    caller.run(inputs, controls)

    outname = args.out_name if args.out_name is not None else ""
    if args.sequence_graph is not None:
        PeakFasta(args.sequence_graph).write_max_path_sequences(
            outname+"sequences.fasta", caller.max_path_peaks)
    else:
        logging.info(
            "Not saving max path sequences, since a sequence graph was not found.")


def find_or_create_linear_map(graph, linear_map_name):

    if os.path.isfile(linear_map_name):
        logging.info("Found linear map %s. "
                     "Will be used in peak calling." % linear_map_name)
    else:
        logging.info("No linear map provided, and none found. Will create now.")
        create_linear_map(graph, linear_map_name)
        logging.info("Done creating linear map")


def _get_file_names(pattern):
    from glob import glob

    if pattern is None:
        return None

    assert isinstance(pattern, list)
    pattern = sorted(pattern)

    if isinstance(pattern, list):
        if len(pattern) > 1:
            return pattern
        else:
            if "*" in pattern[0]:
                return sorted(glob(pattern[0]))

    return sorted(pattern)


def run_callpeaks2(args):
    graphs = _get_file_names(args.graph)
    samples = _get_file_names(args.sample)
    controls = _get_file_names(args.control)
    if controls is None:
        controls = samples

    if len(samples) != len(graphs):
        logging.critical("""Number of sample files must be equal to number of graph files.
                         Found %d graphs and %d sample files.
                         Use --verbose 2 for more details.""" % (len(graphs), len(samples)))
        sys.exit(1)

    logging.debug("Fragment length input: %s" % args.fragment_length)
    logging.info("Sample files: %s" % samples)
    logging.debug("Control files: %s" % samples)
    logging.debug("Graph files: %s" % graphs)

    logging.info("Using graphs: %s " % graphs)
    sequence_graph_file_names = [fn + ".sequences" for fn in graphs]
    logging.info("Will use sequence graphs. %s" % sequence_graph_file_names)
    sequence_retrievers = (obg.SequenceGraph.from_file(fn)
                           for fn in sequence_graph_file_names)

    data_dir = os.path.dirname(graphs[0])
    if data_dir == "":
        data_dir = "./"

    logging.info("Using graphs from data directory %s" % data_dir)
    if len(graphs) > 1:
        # Make custom names for different runs
        print([fn.split(data_dir, 1)[-1] for fn in graphs])
        names = [fn.split(data_dir, 1)[-1].rsplit(".", 1)[0] for fn in graphs]
    else:
        names = [""]

    logging.info("Will use %s as extra experiments names for each run, based on graph file names."
                 "If only running on single graph, this should be empty. " % names)

    linear_map_file_names = []
    for i, graph_file_name in enumerate(graphs):
        linear_map_name = graph_file_name.split(".nobg")[0] + "_linear_map.npz"
        if not os.path.isfile(linear_map_name):
            logging.warning("Did not find linear map for "
                            " for graph %s. Will create." % graph_file_name)
            graph = obg.Graph.from_file(graphs[i])
            create_linear_map(graph, linear_map_name)
        else:
            logging.info(
                "Found linear map %s that will be used." % linear_map_name)
        linear_map_file_names.append(linear_map_name)

    logging.info("Will use linear maps: %s" % linear_map_file_names)

    config = Configuration()
    if args.keep_duplicates == "True":
        config.keep_duplicates = True
        logging.info("Keeping duplicates")

    if args.fragment_length is None:
        logging.info("Fragment length was not specified. Will now"
                     " predict fragment length.")

        min_m = 5 if args.min_fold_enrichment is None else int(args.min_fold_enrichment)
        max_m = 50 if args.max_fold_enrichment is None else int(args.max_fold_enrichment)
        config.fragment_length = int(MultiGraphShiftEstimator.from_files(
            graphs, samples, min_m, max_m).get_estimates())
        logging.info("Estimated fragment length to be %d" % config.fragment_length)
        assert config.fragment_length < 1000, "Suspiciously high fragment length. Probably a bad estimate." \
                                       " Change -m/-M to try to get more paired peaks."
    else:
        config.fragment_length = int(args.fragment_length)
    if args.read_length is None:
        config.read_length = estimate_read_length(samples[0], graphs[0])
        logging.info("Estimated read length to %s" % config.read_length)
    else:
        config.read_length = int(args.read_length)

    if config.fragment_length < config.read_length:
        logging.critical("Fragment length is smaller than read length. Cannot call peaks.")
        sys.exit(1)

    if args.genome_size is not None:
        genome_size = int(args.genome_size)
        config.global_min = int(args.unique_reads) * int(args.fragment_length) / genome_size

        logging.info(
            "Computed min background signal to be %.3f using fragment length %f, "
            " %d unique reads, and genome size %d" % (config.global_min,
                                                      config.fragment_length,
                                                      int(args.unique_reads),
                                                      int(genome_size)))
    else:
        logging.info("Not using min background.")
        config.global_min = None

    out_name = args.out_name if args.out_name is not None else ""
    reporter = Reporter(out_name)
    config.has_control = args.control is not None
    caller = MultipleGraphsCallpeaks(
        names,
        graphs,
        samples,
        controls,
        linear_map_file_names,
        config, reporter,
        sequence_retrievers=sequence_retrievers,
        stop_after_p_values=args.stop_after_p_values == "True",
    )
    caller.run()


def run_callpeaks_whole_genome(args):
    logging.info("Running run_callpeaks_whole_genome")

    config = Configuration()
    if args.keep_duplicates == "True":
        config.keep_duplicates = True
        logging.info("Keeping duplicates")

    logging.info("Running whole genome.")
    chromosomes = args.chromosomes.split(",")
    graph_file_names = [args.data_dir + "/" + chrom + ".nobg" for chrom in chromosomes]
    logging.info("Will use graphs: %s" % graph_file_names)

    linear_map_file_names = []
    for i, chrom in enumerate(chromosomes):
        linear_map_name = args.data_dir + "/" + chrom + "_linear_map.npz"
        if not os.path.isfile(linear_map_name):
            logging.warning("Did not find linear map for "
                            "chromosome %s. Will create." % chrom)
            graph = obg.Graph.from_file(graph_file_names[i])
            create_linear_map(graph, linear_map_name)
        else:
            logging.info("Found linear map %s that will be used." % linear_map_name)
        linear_map_file_names.append(linear_map_name)


    sequence_retrievers = \
            (obg.SequenceGraph.from_file(args.data_dir + "/" + chrom + ".nobg.sequences")
             for chrom in chromosomes)


    if args.sample.endswith(".intervalcollection"):
        sample_file_names = [args.sample.replace("chrom", chrom) for chrom in chromosomes]

    else:
        sample_base_name = args.sample.replace(".json", "_")
        sample_file_names = [sample_base_name + chrom + ".json" for chrom in chromosomes]

    logging.info("Will use input alignments from %s" % sample_file_names)

    if args.control is not None:
        if args.control.endswith(".intervalcollection"):
            control_file_names = [
                args.control_reads_base_name.replace("chrom", chrom)
                for chrom in chromosomes]
        else:
            control_base_name = args.control.replace(".json", "_")
            control_file_names = [control_base_name + chrom + ".json"
                                  for chrom in chromosomes]
    else:
        logging.info("Using input alignments as control")
        control_file_names = sample_file_names.copy()

    config.fragment_length = int(args.fragment_length)
    config.read_length = int(args.read_length)

    if config.fragment_length < config.read_length:
        logging.critical("Fragment length is smaller than read length. Cannot call peaks.")
        sys.exit(1)


    genome_size = int(args.genome_size)
    config.global_min = int(args.unique_reads) * int(args.fragment_length) / genome_size
    logging.info(
        "Computed min background signal to be %.3f using fragment length %f, "
        " %d unique reads, and genome size %d" % (config.global_min,
                                                  config.fragment_length,
                                                  int(args.unique_reads),
                                                  int(genome_size)))

    out_name = args.out_name if args.out_name is not None else ""
    reporter = Reporter(out_name)
    config.has_control = args.control is not None
    caller = MultipleGraphsCallpeaks(
        chromosomes,
        graph_file_names,
        sample_file_names,
        control_file_names,
        linear_map_file_names,
        config, reporter,
        sequence_retrievers=sequence_retrievers,
        stop_after_p_values=args.stop_after_p_values == "True",
    )
    caller.run()


def run_callpeaks_whole_genome_from_p_values(args):
    logging.info("Running whole genome from p-values.")
    chromosome = args.chromosome
    chromosomes = [chromosome]
    graph_file_names = [args.data_dir + chrom + ".nobg" for chrom in chromosomes]
    sequence_retrievers = \
        (obg.SequenceGraph.from_file(args.data_dir + "/" + chrom + ".nobg.sequences") for chrom in chromosomes)
    out_name = args.out_name if args.out_name is not None else ""
    config = Configuration()

    if args.q_threshold is not None:
        config.q_values_threshold = float(args.q_threshold)
        logging.info("Running with q value threshold %.3f" % config.q_values_threshold)
    else:
        logging.info("Q value threshold not set. Running with default 0.05.")

    config.fragment_length = int(args.fragment_length)
    config.read_length = int(args.read_length)
    reporter = Reporter(out_name)
    caller = MultipleGraphsCallpeaks(
        chromosomes,
        graph_file_names,
        None,
        None,
        None,
        config,
        reporter,
        sequence_retrievers=sequence_retrievers,
    )
    caller.create_joined_q_value_mapping()
    caller.run_from_p_values(only_chromosome=chromosome)



import logging
import numpy as np

from .segmentanalyzer import SegmentAnalyzer
from .graphs import DividedLinegraph, DummyTouched
from ..sparsediffs import SparseValues


class HolesCleaner:
    def __init__(self, graph, sparse_values, max_size, touched_nodes=None):
        sparse_values.indices = sparse_values.indices.astype("int")
        self._graph = graph
        self._node_indexes = graph.node_indexes
        self._sparse_values = sparse_values
        self._holes = self.get_holes()
        self._node_ids = self.get_node_ids()
        self._max_size = max_size
        self._kept = []
        self._touched_nodes = touched_nodes
        if touched_nodes is None:
            self._touched_nodes = DummyTouched()

    def get_holes(self):
        start_idx = 0
        if self._sparse_values.values[0] != 0:
            start_idx += 1
        end_idx = self._sparse_values.indices.size
        self._last_node = self._node_indexes.size
        if self._sparse_values.values[-1] == 0:
            self._last_node = np.searchsorted(
                self._node_indexes, self._sparse_values.indices[-1])
            end_idx -= 1
        return self._sparse_values.indices[start_idx:end_idx].reshape(
            (end_idx-start_idx)//2, 2)

    def build_sparse_values(self, holes):
        indices = np.sort(holes.ravel())
        values = np.arange(indices.size) % 2
        if (not indices.size) or indices[0] != 0:
            indices = np.r_[0, indices]
            values = np.r_[1, values]
        if self._sparse_values.values[-1] == 0:
            indices = np.r_[indices, self._sparse_values.indices[-1]]
            values = np.r_[values, 0]
        new_pileup = SparseValues(indices.astype("int"),
                            values.astype("bool"), sanitize=True)
        new_pileup.track_size = self._sparse_values.track_size
        return new_pileup

    def get_node_ids(self):
        node_ids = np.empty_like(self._holes)
        node_ids[:, 0] = np.digitize(self._holes[:, 0], self._node_indexes)
        node_ids[:, 1] = np.digitize(
            self._holes[:, 1], self._node_indexes, True)
        return node_ids

    def _handle_internal(self, internal_holes):
        keep = (internal_holes[:, 1]-internal_holes[:, 0]) > self._max_size
        self._kept_internals = internal_holes[keep]

    def _filter_touched_nodes(self, node_values):
        if not self._touched_nodes:
            return node_values
        touched = [i for i, node in enumerate(node_values[0])
                   if node+self._graph.min_node-1 in self._touched_nodes]
        self._not_touched = np.array(
            [i for i in node_values[0]
             if i+self._graph.min_node-1 not in self._touched_nodes],
            dtype="int")
        return node_values[:, touched]

    def run(self):
        analyzer = SegmentAnalyzer(
            self._holes, self._node_ids, self._node_indexes)
        analyzer.run()
        self._handle_internal(analyzer.internals)
        starts = analyzer.get_ends()
        fulls = self._filter_touched_nodes(analyzer.get_fulls())
        ends = analyzer.get_starts()
        linegraph = DividedLinegraph(
            starts, fulls, ends, self._graph,
            self._last_node, self._touched_nodes)
        mask = linegraph.filter_small(self._max_size)
        self._kept_borders = linegraph.get_masked(mask)
        return self.build_sparse_values(self.build_kept_holes())

    def build_kept_holes(self):
        starts, fulls, ends = self._kept_borders
        fulls = np.r_[fulls, self._not_touched]
        n_starts, n_fulls, n_ends = starts.shape[1], fulls.size, ends.shape[1]
        logging.debug("# %s, %s, %s", n_starts, n_fulls, n_ends)
        n_internals = self._kept_internals.shape[0]
        all_holes = np.empty((n_starts+n_fulls+n_ends+n_internals, 2),
                             dtype="int")
        if all_holes.shape == self._kept_internals.shape:
            return self._kept_internals
        if n_starts:
            all_holes[:n_starts, 0] = self._node_indexes[starts[0]] - starts[1]
            all_holes[:n_starts, 1] = self._node_indexes[starts[0]]
        n_border = n_starts+n_fulls+n_ends
        if fulls.size:
            all_holes[n_starts:n_starts+n_fulls, 0] = self._node_indexes[fulls-1]
            all_holes[n_starts:n_starts+n_fulls, 1] = self._node_indexes[fulls]
        if ends.size:
            all_holes[n_starts+n_fulls:n_border, 0] = self._node_indexes[ends[0]-1]
            all_holes[n_starts+n_fulls:n_border, 1] = self._node_indexes[ends[0]-1] + ends[1]
        all_holes[n_border:] = self._kept_internals
        all_holes.sort(axis=0)
        return all_holes



import numpy as np
import pickle
import logging
from .areas import BinaryContinousAreas
from collections import defaultdict


class ConnectedAreas(Areas):
    def __init__(self, graph, areas=None):
        super(ConnectedAreas, self).__init__(graph, areas)

    def touches_area(self, other_node, start, end):
        graph = self.graph

        if start > 0 and end < graph.node_size(other_node):
            if other_node not in self.areas:
                return False

        if start == 0:
            if not self._is_start_position(other_node, start):
                return True

        if end == self.graph.node_size(other_node):
            if not self._is_end_position(other_node, end):
                return True

        return False

    def __add__(self, other):
        # Adds another connected area to this one
        for node_id, starts_and_ends in other.areas.items():
            self.add_areas_for_node(node_id, starts_and_ends)
        return self

    def contains_interval(self, interval):
        intervals = self.to_simple_intervals()
        overlap = 0
        for internal_interval in intervals:
            overlap += interval.overlap(internal_interval)

        assert overlap <= interval.length()

        if overlap == interval.length():
            return True

        return False

    def n_basepairs(self):
        intervals = self.to_simple_intervals()
        n = 0
        for interval in intervals:
            n += interval.length()
        return n


class BCConnectedAreas(BinaryContinousAreas):
    def __init__(self, graph, id, areas=None):
        super(BCConnectedAreas, self).__init__(graph)
        self.id = id
        for node, startend in areas.items():
            self.add(node, startend[0], startend[1])

    def __hash__(self):
        return self.id

    def touches_area(self, other_node, start, end):
        graph = self.graph

        if start > 0 and end < graph.node_size(other_node):
            if other_node not in self.areas:
                return False

        if start == 0:
            if not self._is_start_position(other_node, start):
                return True

        if end == self.graph.node_size(other_node):
            if not self._is_end_position(other_node, end):
                return True

        return False

    def __add__(self, other):
        # Adds another connected area to this one
        self.merge_with_other(other)
        return self

    def contains_interval(self, interval):
        intervals = self.to_simple_intervals()
        overlap = 0
        for internal_interval in intervals:
            overlap += interval.overlap(internal_interval)

        assert overlap <= interval.length()

        if overlap == interval.length():
            return True

        return False

    def n_basepairs(self):
        intervals = self.to_simple_intervals()
        n = 0
        for interval in intervals:
            n += interval.length()
        return n


class SubgraphCollection(object):

    def __init__(self, graph, subgraphs=None):
        self.graph = graph
        self._id_counter = 0
        if subgraphs is not None:
            self.subgraphs = subgraphs
        else:
            self.subgraphs = []

        self._node_end_index = defaultdict(set)  # Index from node id to subgraphs touching end
        self._node_start_index = defaultdict(set)

    def remove(self, subgraph):
        self.subgraphs.remove(subgraph)

    def add(self, subgraph):
        pass

    @classmethod
    def from_pileup(cls, graph, pileup):
        collection = cls(graph)
        logging.info("Finding valued areas")
        areas = pileup.find_valued_areas(1)
        n = 0
        for node_id, starts_and_ends in areas.items():
            if n % 50000 == 0:
                logging.info("Adding node %d to subgraph" % n)
            n += 1
            for i in range(0, len(starts_and_ends) // 2):
                collection.add_area(node_id,
                                    starts_and_ends[i*2],
                                    starts_and_ends[i*2+1])
        return collection

    def _subgraphs_touching_area(self, node_id, start, end):
        out = set()
        if start == 0:
            #print("   Checking start of node %d" % node_id)
            out = self._node_start_index[node_id]
            #print("   Found:")
            #print(self._node_start_index[node_id])
        if end == self.graph.node_size(node_id):
            out.update(self._node_end_index[node_id])
            #print("  Checking end of node %d" % node_id)
            #print("   Found:")
            #print(self._node_end_index[node_id])
        #print("   Found in total %d touching subgraphs" % len(out))
        #return list(set(out))
        return out

        touching_subgraphs = []
        for subgraph in self.subgraphs:
            if subgraph.touches_area(node_id, start, end):
                touching_subgraphs.append(subgraph)
        return touching_subgraphs

    def __iter__(self):
        return iter(self.subgraphs)

    def add_indexes(self, node, start, end, subgraph):
        # Checks if added node id,start,end touches nodes. Link index to subgraph
        if start == 0:
            for in_node in self.graph.adj_list[-node]:
                #print("    Adding subgraph to end index of %d and start of %d" % (in_node, -in_node))
                self._node_end_index[in_node].add(subgraph)
                self._node_start_index[-in_node].add(subgraph)
            for in_node in self.graph.reverse_adj_list[-node]:
                #print("    Adding subgraph to end index of %d and end of %d" % (-in_node, in_node))
                self._node_end_index[-in_node].add(subgraph)
                self._node_start_index[in_node].add(subgraph)

        if end == self.graph.node_size(node):
            for in_node in self.graph.adj_list[node]:
                #print("    Adding subgraph to start index of %d and end index of %d" % (in_node, -in_node))
                self._node_start_index[in_node].add(subgraph)
                self._node_end_index[-in_node].add(subgraph)
            for in_node in self.graph.reverse_adj_list[node]:
                #print("    Adding subgraph to end index of %d and start of %d" % (-in_node, in_node))
                self._node_end_index[-in_node].add(subgraph)
                self._node_start_index[in_node].add(subgraph)



    def add_area(self, node_id, start, end):
        assert node_id in self.graph.blocks
        assert start >= 0 and start < end
        assert end <= self.graph.node_size(node_id)
        #print("Adding node %d, startend %d %d" % (node_id, start, end))
        #print("%d subgraphs so far" % len(self.subgraphs))
        if start > 0 and end < self.graph.node_size(node_id):
            # Internal, will never touch anything
            touching_subgraphs = []
        else:
            touching_subgraphs = list(self._subgraphs_touching_area(node_id, start, end))

        self._id_counter += 1
        if len(touching_subgraphs) == 0:
            new_subgraph = BCConnectedAreas(
                self.graph,
                self._id_counter,
                {node_id: np.array([start, end])})
            self.subgraphs.append(new_subgraph)
            self.add_indexes(node_id, start, end, new_subgraph)

        elif len(touching_subgraphs) == 1:
            touching_subgraphs[0].add(
                node_id, start, end)
            #print("Subgraph after adding")
            #print(touching_subgraphs[0])
            self.add_indexes(node_id, start, end, touching_subgraphs[0])

        elif len(touching_subgraphs) > 1:
            #print("  Found multiple subgraphs")
            # Merge all touching subgraphs, then add the area
            new_subgraph = touching_subgraphs[0]
            for touching_subgraph in touching_subgraphs[1:]:
                new_subgraph = new_subgraph + touching_subgraph
                if touching_subgraph in self.subgraphs:
                    self.remove(touching_subgraph)

            new_subgraph.add(node_id, start, end)
            self.add_indexes(node_id, start, end, new_subgraph)

    def to_file(self, file_name):
        f = open(file_name, "w")
        lines = (subgraph.to_file_line() for subgraph in self.subgraphs)
        f.writelines(lines)
        f.close()

    def contains_interval(self, interval):
        for connected_area in self.subgraphs:
            if connected_area.contains_interval(interval):
                return True

        return False

    @classmethod
    def from_pickle(cls, file_name, graph=None):
        with open("%s" % file_name, "rb") as f:
            obj = pickle.loads(f.read())
            assert isinstance(obj, cls)

            for subgraph in obj.subgraphs:
                subgraph.graph = graph

            return obj

    def to_pickle(self, file_name):
        with open("%s" % file_name, "wb") as f:
            pickle.dump(self, f)


class SubgraphCollectionPartiallyOrderedGraph(SubgraphCollection):
    @classmethod
    def create_from_pileup(cls, graph, pileup):
        builder = SubgraphCollectionPartiallyOrderedGraphBuilder(
                    graph, pileup)
        return builder.build()


class SingleArea:
    def __init__(self, node, start, end, is_start, is_end):
        self.node = node
        self.start = start
        self.end = end
        self.is_start = is_start
        self.is_end = is_end

    def __str__(self):
        return "Node %d, start/end: %d/%d. Is start: %s, is end:%s" % \
               (self.node, self.start, self.end, self.is_start, self.is_end)

    def __repr__(self):
        return self.__str__()


class SubgraphCollectionPartiallyOrderedGraphBuilder():

    def __init__(self, graph, pileup):
        self.graph = graph
        self.pileup = pileup
        from .densepileup import DensePileupData
        assert isinstance(self.pileup.data, DensePileupData)
        areas = pileup.find_valued_areas(1)
        self.areas = Areas(self.graph, areas)
        self.single_areas = []

    def _make_single_areas(self):
        logging.info("Making single areas")
        sorted_nodes = sorted(self.areas.areas.keys())
        n = 0
        for node_id in sorted_nodes: #, starts_and_ends in self.areas.areas.items():
            #print("Checking node %d" % node_id)
            starts_and_ends = self.areas.areas[node_id]
            if n % 1000000 == 0:
                logging.info("Adding node %d to subgraph" % n)
            n += 1
            for i in range(0, len(starts_and_ends) // 2):
                start = starts_and_ends[i*2]
                end = starts_and_ends[i*2+1]
                is_start = self.areas._is_start_position(node_id, start)
                #print("Checking end %d, %d" % (node_id, end))
                is_end = self.areas._is_end_position(node_id, end)
                yield SingleArea(node_id, start, end, is_start, is_end)

    def _get_connected_subgraphs(self, subgraphs, area):
        connected = []
        #print("Getting connected subgraphs")
        for subgraph in subgraphs:
            #print("  Checking subgraph")
            if not area.is_start:
                for prev_node in self.graph.reverse_adj_list[-area.node]:
                    #print("        Checking prev node %d" % prev_node)
                    if -prev_node in subgraph.full_areas:
                        connected.append(subgraph)
                    if prev_node in subgraph.starts:
                        connected.append(subgraph)
            elif not area.is_end:
                for next_node in self.graph.adj_list[area.node]:
                    #print("        Checking next node %d" % next_node)
                    if next_node in subgraph.starts:
                        connected.append(subgraph)
                    if next_node in subgraph.full_areas:
                        connected.append(subgraph)

        return connected

    def _create_subgraphs(self):
        from collections import deque
        active_subgraphs = deque()

        logging.info("Creating subgraphs after single areas are created")
        peaks = []
        current_peak = BinaryContinousAreas(self.graph)
        prev_was_end = False
        prev_node = 0
        for single_area in self.single_areas:
            #print("Single area: %s " % single_area)

            connected = self._get_connected_subgraphs(active_subgraphs, single_area)
            if len(connected) > 1:
                merged = connected[0]
                for subgraph in connected[1:]:
                    merged.merge_with_other(subgraph)

                connected = [merged]

            if len(connected) == 0:
                current_peak = BinaryContinousAreas(self.graph)
                current_peak.add(single_area.node, single_area.start,
                                 single_area.end)
                if single_area.is_start and single_area.is_end:
                    peaks.append(current_peak)
                    #print("   Creating new, finishing it")
                else:
                    active_subgraphs.append(current_peak)
                    #print("   Creating new, appending to active")
            else:
                #print("   Adding to active")
                current_peak = connected[0]

                current_peak.add(single_area.node,
                             single_area.start,
                             single_area.end)

            # If more than 2, the first must be finished
            if len(active_subgraphs) > 5:
                peaks.append(active_subgraphs.popleft())
                #print("Finishing subgraph")

        peaks.extend(active_subgraphs)

        return peaks


    def build(self):
        self.single_areas = self._make_single_areas()
        return self._create_subgraphs()

    """
    def build(self):
        sorted_nodes = sorted(self.areas.keys())

        starts = OrderedDict()
        ends = OrderedDict()

        for node in sorted_nodes:
            starts_ends = self.areas[node]
            assert len(starts_ends) == 2 or len(starts_ends) == 4

            if self.areas._is_start_position(node, starts_ends[0]):
                starts[node] = starts_ends[:2]

            if self.areas._is_end_position(node, starts_ends[-1]):
                ends[node] = starts_ends[-2:]


        # Iterate sorted nodes. Count number of starts and number of ends
        in_peak = False
        prev_status = in_peak
        peaks = []
        current_peak = BinaryContinousAreas(self.graph)
        for node in sorted_nodes:
            start_ends = self.areas[node]
            assert len(start_ends) == 2 or len(starts_ends) == 4
            is_start = node in starts
            is_end = node in ends

            if is_start and is_end:
                current_peak.add(node, start_ends[0:2])
    """



import numpy as np
from operator import itemgetter


class EventSorter:

    def np_init(self, index_lists, values_lists):
        all_indices = np.concatenate(index_lists)
        all_values = np.concatenate(values_lists)
        codes = np.empty(all_indices.shape, dtype="uint8")
        start = 0
        for i, index_list in enumerate(index_lists):
            codes[start:start+len(index_list)] = i
            start += len(index_list)
        args = np.lexsort((codes, all_indices))
        self.values = all_values[args]
        self.indices = all_indices[args]
        self.codes = codes[args]
        self.n_codes = len(index_lists)

    def __init__(self, index_lists, values_lists, names=None):

        if names is not None:
            [setattr(self, name.upper(), i) for i, name in enumerate(names)]

        return self.np_init(index_lists, values_lists)

    def __str__(self):
        return "EventSorter(\n" +"\n".join(str(t) for t in self) + ")"

    def __iter__(self):
        # return self.tuples.__iter__()
        return zip(self.indices, self.codes, self.values)


class EventSort(object):
    def __init__(self, index_lists, codes, names=None):
        if names is not None:
            [setattr(self, name.upper(), i) for i, name in enumerate(names)]
        self.tuples = []
        for index_list, code in zip(index_lists, codes):
            self.tuples.extend([(idx, code) for idx in index_list])
        self.tuples.sort(key=itemgetter(0, 1))
        values = np.array([t[1] for t in self.tuples])
        indices = np.array([t[0] for t in self.tuples])
        values = np.cumsum(values)
        new_values = np.nonzero(np.diff(values))[0] + 1
        self.values = np.empty(new_values.size+1)
        self.indices = np.empty(new_values.size+1, dtype="int")

        self.indices[1:] = indices[new_values]
        self.indices[0] = indices[0]

        self.values[1:] = values[new_values]
        self.values[0] = values[0]



from ..linear_filter import LinearFilter
from .shiftestimation import Opt, Treatment, PeakModel
import logging


class MultiGraphShiftEstimator(object):
    def __init__(self, position_tuples, genome_size,
                 min_fold_enrichment=5, max_fold_enrichment=50):
        self._position_tuples = position_tuples
        self.genome_size = genome_size
        self.min_fold_enrichment = min_fold_enrichment
        self.max_fold_enrichment = max_fold_enrichment

    def get_estimates(self):
        opt = Opt(self.min_fold_enrichment, self.max_fold_enrichment)
        opt.gsize = self.genome_size
        treatment = Treatment(self._position_tuples)
        peakmodel = PeakModel(opt, treatment)
        peakmodel.build()
        return round(peakmodel.d)

    @classmethod
    def from_files(cls, graph_file_names, interval_json_file_names, min_fold_enrichment=5,
                   max_fold_enrichment=50):
        start_positions = {
            "+": {str(i): [] for i, _ in enumerate(graph_file_names)},
            "-": {str(i): [] for i, _ in enumerate(graph_file_names)}
        }
        genome_size = 0
        i = 0
        for graph, intervals in zip(graph_file_names,
                                    interval_json_file_names):
            linear_filter = LinearFilter.from_vg_json_reads_and_graph(
                intervals, graph)

            positions = linear_filter.find_start_positions()
            start_positions["+"][str(i)] = positions["+"]
            start_positions["-"][str(i)] = positions["-"]
            i += 1

            genome_size += linear_filter._indexed_interval.length()

        logging.info("Using genome size %d" % genome_size)
        return cls(start_positions, genome_size, min_fold_enrichment, max_fold_enrichment)

    @classmethod
    def _from_files(cls, graph_file_names, interval_json_file_names):
        start_positions = {
            "+": {name: [] for name in chrom_names},
            "-": {name: [] for name in chrom_names}
        }
        genome_size = 0

        for name, graph, intervals in zip(chrom_names,
                                          graph_file_names,
                                          interval_json_file_names):

            linear_filter = LinearFilter.from_vg_json_reads_and_graph(
                intervals, graph)

            positions = linear_filter.find_start_positions()
            start_positions["+"][name] = positions["+"]
            start_positions["-"][name] = positions["-"]

            genome_size += linear_filter._indexed_interval.length()

        logging.info("Using genome size %d" % genome_size)
        return cls(start_positions, genome_size)



import logging
from pyvg.conversion import vg_json_file_to_interval_collection,\
    json_file_to_obg_numpy_graph
import offsetbasedgraph as obg
from graph_peak_caller.shiftestimation.shift_estimation_multigraph \
    import MultiGraphShiftEstimator
from graph_peak_caller.util import create_linear_map
from graph_peak_caller.multiplegraphscallpeaks import MultipleGraphsCallpeaks


def count_unique_reads_interface(args):
    chromosomes = args.chromosomes.split(",")
    graph_file_names = [args.graphs_location + chrom + ".nobg" for chrom in chromosomes]
    reads_file_names = [args.reads_base_name + chrom + ".json"
                        for chrom in chromosomes]

    count_unique_reads(chromosomes, graph_file_names,
                       reads_file_names)


def shift_estimation(args):
    chromosomes = args.chromosomes.split(",")
    graphs = [args.ob_graphs_location + "/" + chrom + ".nobg"
              for chrom in chromosomes]
    logging.info("Will try to use graphs %s" % graphs)
    sample_file_names = [args.sample_reads_base_name + chrom + ".json"
                         for chrom in chromosomes]
    logging.info("Will use reads from %s" % sample_file_names)

    min_m = 5 if args.min_fold_enrichment is None else int(args.min_fold_enrichment)
    max_m = 50 if args.max_fold_enrichment is None else int(args.max_fold_enrichment)

    estimator = MultiGraphShiftEstimator.from_files(graphs, sample_file_names,
                                                    min_fold_enrichment=min_m,
                                                    max_fold_enrichment=max_m)

    d = estimator.get_estimates()
    logging.info("Found shift: %d" % d)


def count_unique_reads(chromosomes, graph_file_names, reads_file_names):
    graphs = (obg.GraphWithReversals.from_numpy_file(f)
              for f in graph_file_names)
    reads = (vg_json_file_to_interval_collection(f, graph)
             for f, graph in zip(reads_file_names, graphs))

    unique_reads = MultipleGraphsCallpeaks.count_number_of_unique_reads(reads)
    print(unique_reads)


def create_ob_graph(args):
    logging.info("Creating obgraph")
    ob_graph = json_file_to_obg_numpy_graph(args.vg_json_file_name, 0)
    logging.info("Writing ob graph to file")
    out_name = args.out_file_name if args.out_file_name is not None \
                else args.vg_json_file_name.split(".")[0] + ".nobg"
    ob_graph.to_numpy_file(out_name)

    logging.info("Creating sequence graph")
    from offsetbasedgraph import SequenceGraph
    sequence_graph = SequenceGraph.create_empty_from_ob_graph(ob_graph)
    sequence_graph.set_sequences_using_vg_json_graph(args.vg_json_file_name)
    out_file_name = out_name + ".sequences"
    sequence_graph.to_file(out_file_name)
    logging.info("Wrote sequence graph to %s" % out_file_name)


def create_linear_map_interface(args):
    graph = args.graph
    # graph.convert_to_dict_backend()

    out_name = args.out_file_base_name if args.out_file_base_name is not None \
        else args.graph_file_name.split(".")[0] + "_linear_map.npz"
    create_linear_map(graph, out_name)
    logging.info("Wrote linear map to file %s" % out_name)


def split_vg_json_reads_into_chromosomes(args):
    reads_base_name = args.vg_json_reads_file_name.split(".")[0]
    logging.info("Will write reads to files %s_[chromosome].json",
                 reads_base_name)

    chromosomes = args.chromosomes.split(",")
    chromosome_limits = {}
    logging.info("Found the following chromosome ranges:")
    for chrom in chromosomes:
        f = open(
            args.range_files_base_name + "node_range_" + chrom + ".txt")
        start_end = f.read().split(":")
        start = int(start_end[0])
        end = int(start_end[1])
        chromosome_limits[chrom] = (start, end)
        logging.info("   Chr%s: %d-%d" % (chrom, start, end))
        f.close()

    out_files = {chrom: open(reads_base_name + "_" + chrom + ".json", "w")
                 for chrom in chromosomes}

    reads_file = open(args.vg_json_reads_file_name)
    i = 0
    import re
    regex = re.compile(r"node_id\":\s?\"?([0-9]+)\"?")

    def get_mapped_chrom(node):
        mapped_chrom = None
        for chrom in chromosomes:
            if node >= chromosome_limits[chrom][0] and node <= chromosome_limits[chrom][1]:
                mapped_chrom = chrom
                break
        return mapped_chrom

    n_without_node_id = 0
    for line in reads_file:
        if i % 100000 == 0:
            logging.info("Line #%d" % i)
        i += 1

        groups = regex.search(line)
        if groups is None:
            n_without_node_id += 1
            continue
        groups = groups.groups()
        if len(groups) > 0:
            node = int(groups[0])
            mapped_chrom = get_mapped_chrom(node)
            if mapped_chrom is None:
                n_without_node_id += 1
                continue
            out_files[mapped_chrom].writelines([line])
        else:
            print("No groups fond")

    for file in out_files.values():
        file.close()

    logging.info("Done. Found %d lines without node id or not matching into the given list of chromosomes" % n_without_node_id)



import logging
from collections import defaultdict
from offsetbasedgraph.graphtraverser import GraphTraverser
from offsetbasedgraph.interval import Position
from .areas import BinaryContinousAreas
import offsetbasedgraph as obg
import numpy as np
import itertools


class Areas(object):
    def __init__(self, graph, areas=None):
        self.graph = graph
        if areas is None:
            self.areas = {}
        else:
            self.areas = areas

    def __eq__(self, other):
        if not len(self.areas) == len(other.areas):
            return False

        if not all(node_id in other.areas for node_id in self.areas):
            return False

        return all(startend == other.areas[node_id] for node_id, startend
                   in self.areas.items())

    def __str__(self):
        return str(self.areas)

    def __repr__(self):
        return repr(self.areas)

    def join(self, areas):
        pass

    @classmethod
    def from_interval(cls, interval, graph):
        if len(interval.region_paths) == 1:
            return cls(graph, {interval.region_paths[0]: [
                interval.start_position.offset,
                interval.end_position.offset]})
        areas = {}
        start_rp = interval.region_paths[0]
        start_len = graph.node_size(start_rp)-interval.start_position.offset
        areas[-start_rp] = [0, start_len]
        areas[interval.end_position.region_path_id] = [0, interval.end_position.offset]

        for region_path in interval.region_paths[1:-1]:
            areas[region_path] = [0, graph.node_size(region_path)]

        return cls(graph, areas)

    def update(self, other):
        """TODO: Hanlde cyclic graphs"""
        for node_id, startend in other.areas.items():
            if node_id not in self.areas:
                self.areas[node_id] = startend
                continue
            self.areas[node_id] = [
                min(startend[0], self.areas[node_id][0]),
                max(startend[1], self.areas[node_id][1])]

    def _add_nontrivial_areas_for_node(self, node_id, starts_and_ends):
        raise NotImplementedError("Nontrivial addition not supported")

    def add_areas_for_node(self, node_id, starts_and_ends):
        if node_id not in self.areas:
            self.areas[node_id] = starts_and_ends
        else:
            current = self.areas[node_id]
            if len(current) > 2 or len(starts_and_ends) > 2:
                return self._add_nontrivial_areas_for_node(
                    node_id, starts_and_ends)

            if current[1] < starts_and_ends[0]:
                new = np.append(current, starts_and_ends)
            elif current[0] > starts_and_ends[1]:
                new = np.append(starts_and_ends, current)
            else:
                # todo
                print("Error")
                print(current)
                print(starts_and_ends)
                raise NotImplementedError(
                    "Case where new start equals old end is not implemented")

            assert len(new) == 4
            self.areas[node_id] = new

    def robust_update(self, other):
        for node_id, startend in other.areas.items():
            if node_id not in self.areas:
                self.areas[node_id] = startend
                continue

            self.areas[node_id] = [
                min(startend[0], self.areas[node_id][0]),
                max(startend[1], self.areas[node_id][1])]

    def reverse_reversals(self):
        neg_node_ids = [node_id for node_id in self.areas.keys()
                        if node_id < 0]
        for node_id in neg_node_ids:
            
            startend = self.areas[node_id]
            l = self.graph.node_size(node_id)
            pos_coords = [l-pos for pos in reversed(startend)]
            del self.areas[node_id]
            if -node_id not in self.areas:
                self.areas[-node_id] = pos_coords
                continue
            forward_areas = self.areas[-node_id]

            if forward_areas[-1] < pos_coords[0]:
                forward_areas.extend(pos_coords)
            else:
                all_coords = forward_areas + pos_coords
                self.areas[-node_id] = [min(all_coords), max(all_coords)]

    def get_starts(self, node_id):
        return [idx for i, idx in enumerate(self.areas[node_id])
                if i % 2 == 0]

    def get_ends(self, node_id):
        return [idx for i, idx in enumerate(self.areas[node_id])
                if i % 2 == 1]

    def to_intervals(self, include_partial_stubs):
        raise NotImplementedError()

    def to_simple_intervals(self):
        intervals = []
        for node, area in self.areas.items():
            for i in range(0, len(area) // 2):
                start = int(area[i*2])
                end = int(area[i*2+1])
                interval = obg.Interval(start, end, [node], self.graph)
                intervals.append(interval)

        return intervals

    def _is_start_or_end_position(self, node, position_offset):
        if self._is_end_position(node, position_offset) \
                or self._is_start_position(node, position_offset):
            return True

        return False

    def _is_end_position(self, node, offset):
        if offset < self.graph.node_size(node):
            return True
        graph = self.graph
        for in_node in itertools.chain(graph.adj_list[node], graph.reverse_adj_list[node]):
            if in_node in self.areas and len(self.areas[in_node]) > 0:
                if self.areas[in_node][0] == 0:
                    return False
            elif -in_node in self.areas and len(self.areas[-in_node]) > 0:
                if self.areas[-in_node][-1] == graph.node_size(in_node):
                    return False

        return True

    def _is_start_position(self, node, offset):
        if offset > 0:
            return True

        graph = self.graph
        for in_node in itertools.chain(graph.adj_list[-node], graph.reverse_adj_list[-node]):
            if in_node in self.areas and len(self.areas[in_node]) > 0:
                if self.areas[in_node][0] == 0:
                    return False

            elif -in_node in self.areas and len(self.areas[-in_node]) > 0:
                if self.areas[-in_node][-1] == graph.node_size(in_node):
                    return False

        return True

    def get_start_and_end_positions(self):
        positions = []
        for node, starts_and_ends in self.areas.items():
            n_starts_ends = len(starts_and_ends)
            for start in starts_and_ends[range(0, n_starts_ends, 2)]:
                if self._is_start_position(node, start):
                    positions.append(obg.Position(node, start))

            for end in starts_and_ends[range(1, n_starts_ends, 2)]:
                if self._is_end_position(node, end):
                    positions.append(obg.Position(node, end-1))
                    # Correct to non-inclusive end, since this may be interpreted as a start

        return positions

    def get_all_included_nodes(self):
        return self.areas.keys()

    def to_file_line(self):
        start_ends = ','.join([str(position) for position in
                               self.get_start_and_end_positions()])
        nodes = ','.join([str(node) for node in self.get_all_included_nodes()])
        return "%s\t%s\n" % (start_ends, nodes)

    @classmethod
    def from_file_line(cls, line):
        pass


class AreasBuilder(object):
    def __init__(self, graph):
        self.graph = graph
        self.areas = {}

    def update(self, new_areas):
        """NB!: Only works if all intervals start at 0"""
        for rp, startend in new_areas.items():
            if rp not in self.areas:
                self.areas[rp] = startend
                continue
            self.areas[rp][-1] = max(startend[-1], self.areas[rp][-1])

    def reverse_reversals(self):
        neg_rps = [rp for rp in self.areas if rp < 0]
        for rp in neg_rps:
            node_size = self.graph.node_size(rp)
            pos_rp = -rp
            if pos_rp not in self.areas:
                self.areas[pos_rp] = [
                    node_size-i for i in self.areas[rp][::-1]]
            else:
                mid = node_size-self.areas[rp][-1]
                if self.areas[pos_rp][-1] > mid:
                    self.areas[pos_rp] = [0, node_size]
                else:
                    self.areas[pos_rp] += [mid, node_size]
            del self.areas[rp]

    def filled_interval(self, interval, fill_pos=0, fill_neg=0):
        start = max(interval.start_position.offset-fill_neg, 0)
        end_size = self.graph.node_size(interval.end_position.region_path_id)
        end = min(interval.end_position.offset+fill_pos,
                  end_size)
        neg_remain = max(0, fill_neg-interval.start_position.offset)
        pos_remain = max(0, fill_pos-(end_size-interval.end_position.offset))

        if len(interval.region_paths) == 1:
            # Make sure it touches edge if possible
            if start == 0:
                self.areas[interval.region_paths[0]] = [start, end]
            else:
                self.areas[-interval.region_paths[0]] = [end_size-end, end_size-start]
            return pos_remain, neg_remain
        start_rp = interval.region_paths[0]
        start_len = self.graph.node_size(start_rp) - start
        self.areas[-start_rp] = [0, start_len]
        self.areas[interval.end_position.region_path_id] = [0, end]

        for region_path in interval.region_paths[1:-1]:
            self.areas[region_path] = [0, self.graph.node_size(region_path)]

        return pos_remain, neg_remain


class Extender(object):
    def __init__(self, graph, length):
        self.length = length
        assert isinstance(length, int)
        self.graph = graph
        self.pos_traverser = GraphTraverser(graph, +1)
        self.neg_traverser = GraphTraverser(graph, -1)

    def get_areas_from_node(self, region_path, length, traverser):
        visited = defaultdict(int)
        for next_node in traverser.adj_list[region_path]:
            traverser.extend_from_block(next_node, length, visited)

        for node_id, l in visited.items():
            if l == 0:
                continue
            if l >= self.graph.node_size(node_id):
                self.area_builder.add_full(node_id)
            else:
                self.area_builder.add_start(node_id, l)

    def extend_interval(self, interval, direction=1):
        self.area_builder = BinaryContinousAreas(self.graph)
        pos_length = self.length-interval.length()
        neg_length = self.length if direction == 0 else 0
        pos_remain, neg_remain = self.area_builder.filled_interval(
            interval, pos_length, neg_length)
        is_pos = interval.can_be_on_positive_strand()
        is_neg = interval.can_be_on_negative_strand()
        #print(" Extending interval %s. Can be on pos/neg: %d/%d" % (interval, is_pos, is_neg))
        if pos_remain:
            if is_pos:
                self.get_areas_from_node(
                    interval.end_position.region_path_id,
                    pos_remain, self.pos_traverser)
            if is_neg:
                self.get_areas_from_node(
                    interval.end_position.region_path_id,
                    pos_remain, self.neg_traverser)
        if neg_remain:
            if is_pos:
                self.get_areas_from_node(
                    -interval.start_position.region_path_id,
                    neg_remain, self.neg_traverser)
            if is_neg:
                self.get_areas_from_node(
                    -interval.start_position.region_path_id,
                    neg_remain, self.pos_traverser)

        self.area_builder.sanitize()
        # self.area_builder.reverse_reversals()
        return self.area_builder
# Areas(self.graph, self.area_builder.areas)



import logging
from .control.linearmap import LinearMap


def create_linear_map(ob_graph, out_file_name="linear_map.npz"):
    linear_map = LinearMap.from_graph(ob_graph)
    linear_map.to_file(out_file_name)
    logging.info("Created linear map, wrote to file %s" % out_file_name)



from collections import defaultdict


class MotifLocations:
    def __init__(self, peaks, motifs):
        self._peaks = {peak.unique_id: peak for peak in peaks}
        self._motifs = motifs
        self._locations = self._get_motif_locations()

    def _get_motif_locations(self):
        locations = []
        for peak_id, interval in self._peaks.items():
            motif_locations = self._motifs.get_entries(peak_id)
            for motif_location in motif_locations:
                locations.append(
                    interval.get_subinterval(motif_location._start,
                                             motif_location._end))
        return locations

    def get_rp_index(self):
        index = defaultdict(list)
        for location in self._locations:
            for rp in location.region_paths:
                index[rp].append(location)

    def find_matches(self, other):
        index = self.get_rp_index()
        match_dict = {}
        for location in other._locations:
            matches = []
            for rp in location.region_paths:
                matches.extend(index[rp])
            match_dict[location] = matches
        return match_dict



import numpy as np
from pyfaidx import Fasta
import offsetbasedgraph as obg
import logging
from collections import namedtuple, deque
from itertools import chain


Variant = namedtuple("Variant", ["offset", "ref", "alt"])
FullVariant = namedtuple("FullVariant", ["offset", "ref", "alt", "precence"])


def find_valid_haplotypes(variants, combination):
    combination = np.asanyarray(combination)
    if not np.any(combination):
        return np.arange(1024)
    nonzero = np.flatnonzero(combination)
    informative_haplotypes = []
    informative_variants = [variant for i, variant in enumerate(variants)
                            if i in nonzero]
    for i, variant in enumerate(informative_variants):
        informative_haplotypes.extend(
            variant.precence.get_samples(combination[nonzero[i]]))
    f = np.unique(informative_haplotypes)
    if not f.size:
        return np.array([])
    for code, variant in zip(combination, variants):
        f = variant.precence.get_samples(code, f)
    return f


class VariantPrecence:
    strict = False
    def __init__(self, precence):
        self._precence = precence

    def get_samples(self, variant, f=None):
        s = self._precence if f is None else self._precence[f, :]
        if self.accept_ref and variant == 0:
            return f if f is not None else np.arange(self._precence.shape[0])
        match = np.any((s == variant) | (s == -1),
                       axis=-1)
        res = np.flatnonzero(match)
        if f is not None:
            return f[res]
        return res

    @classmethod
    def from_line(cls, line):
        if cls.strict:
            parts = line.replace(".", "0").replace("/", "|").split("\t")
        else:
            parts = line.replace(".", "-1").replace("/", "|").split("\t")
        haplo_type = [part.split(":", 2)[0].split("|") for part in parts]
        precence = np.array([[int(a), int(b)] for a, b in haplo_type])
        return cls(precence)

    @classmethod
    def join(cls, precences, counts):
        offsets = [0] + list(np.cumsum(counts))[:-1]
        combined = np.zeros_like(precences[0]._precence)
        for precence in precences:
            combined[precence._precence < 0] = -1
        for precence, offset in zip(precences, offsets):
            combined[precence._precence > 0] = precence._precence[precence._precence > 0] + offset
        return cls(combined)

    def __repr__(self):
        return "P(%s)" % np.count_nonzero(self._precence > 0)


class VariantList:
    def __init__(self, start, end):
        self._start = int(start)
        self._end = int(end)
        self._variants = []

    def append(self, variant):
        self._variants.append(variant)

    def finalize(self):
        return [FullVariant(variant.offset-self._start,
                            variant.ref, variant.alt,
                            variant.precence) for variant in self._variants]

        final_list = []
        cur_end = 0
        var_buffer = []

        for variant in self._variants:
            transformed_var = FullVariant(
                variant.offset-self._start,
                variant.ref, variant.alt,
                variant.precence)
            assert transformed_var.offset <= self._end-self._start, (variant, self._start, transformed_var.offset, self._end-self._start)
            if transformed_var.offset >= cur_end:
                if var_buffer:
                    final_list.append(self._join_variants(var_buffer))
                var_buffer = []
            var_buffer.append(transformed_var)
            cur_end = max(transformed_var.offset + len(transformed_var.ref),
                          cur_end)
        if var_buffer:
            final_list.append(self._join_variants(var_buffer))
        return final_list

    def _join_variants(self, variants):
        if len(variants) == 1:
            return variants[0]
        start = min(variant.offset for variant in variants)
        end = max(variant.offset + len(variant.ref) for variant in variants)
        ref = []*(end-start)
        for variant in variants:
            ref[variant.offset-start:variant.offset-start+len(variant.ref)] = variant.ref
        ref = "".join(ref)
        alts = [ref[:variant.offset-start] + v_alt + ref[variant.offset+len(variant.ref)-start:]
                for variant in variants for v_alt in variant.alt]
        counts = [len(variant.alt) for variant in variants]
        precences = [variant.precence for variant in variants]
        combined_precence = VariantPrecence.join(precences, counts)
        return FullVariant(start, ref, alts, combined_precence)


class VCF:
    def __init__(self, vcf_file_name):
        self.file_name = vcf_file_name
        self.f = open(self.file_name)
        self.cur_lines = []

    def _join_variants(self, variants):
        if len(variants) == 1:
            return variants[0]
        start = min(variant.offset for variant in variants)
        end = max(variant.offset + len(variant.ref) for variant in variants)
        ref = []*(end-start)
        for variant in variants:
            ref[variant.offset-start:variant.offset-start+len(variant.ref)] = variant.ref
        ref = "".join(ref)
        alts = [ref[:variant.offset-start] + v_alt + ref[variant.offset+len(variant.ref)-start:]
                for variant in variants for v_alt in variant.alt]
        return Variant(start, ref, alts)

    def _prune_seqs(self, ref, alts):
        offset = 0
        for cs in zip(*([ref]+alts)):
            if all(c == cs[0] for c in cs):
                offset += 1
            else:
                break
        return ref[offset:], [alt[offset:] for alt in alts], offset

    def get_variants_from_intervals(self, intervals):
        """ intervals reveresely sorted on start """
        PRUNE = True
        StackElement = namedtuple("StackElement", ["end", "variant_list"])
        intervals = iter(intervals)
        current_intervals = deque([])
        cur_interval = next(intervals)
        is_finished = False
        line_parts = (line.split("\t", 9) for line in self.f
                      if not line.startswith("#"))
        for parts in line_parts:
            pos = int(parts[1])-1
            while pos >= cur_interval[0] and not is_finished:
                current_intervals.append(
                    StackElement(cur_interval[-1],
                                 VariantList(*cur_interval)))
                try:
                    cur_interval = next(intervals)
                except StopIteration:
                    is_finished = True
            if not current_intervals:
                continue

            while current_intervals and current_intervals[0].end < pos:
                yield current_intervals.popleft().variant_list.finalize()
            if not current_intervals:
                if is_finished:
                    break
                continue
            ref = parts[3].lower()
            alt = parts[4].lower().split(",")
            if PRUNE:
                ref, alt, offset = self._prune_seqs(ref, alt)
                pos = int(pos) + offset
            precence = VariantPrecence.from_line(parts[-1])
            var = FullVariant(int(pos), ref.lower(), alt, precence)
            variant_end = pos+len(ref)
            for stack_element in current_intervals:
                if stack_element.end > variant_end:
                    stack_element.variant_list.append(var)

    def get_variants_from(self, start, end):
        variants = []
        cur_end = 0
        intersecting = []
        for line in chain(self.cur_lines, self.f):
            if line.startswith("#"):
                continue
            parts = line.split("\t")
            pos = int(parts[1])-1
            if pos < start:
                continue
            if pos >= cur_end:
                if intersecting:
                    variants.append(self._join_variants(intersecting))
                intersecting = []
            if pos >= end:
                self.cur_lines = [line]
                break

            ref = parts[3]
            intersecting.append(Variant(int(pos-start), ref, parts[4].split(",")))
            cur_end = max(pos+len(ref), cur_end)
            continue
        if intersecting:
            variants.append(self._join_variants(intersecting))

        self.cur_lines = []
        variants.sort(key=lambda x: x.offset)
        return variants


def traverse_variants(alt_seq, ref_seq, variants):
    if not len(variants):
        return None
    Combination = namedtuple("Combination", ["code", "alt_offset", "prev_offset"])
    combinations = [Combination([], 0, 0)]
    for j, variant in enumerate(variants):
        new_combinations = []
        for comb in combinations:
            inter_alt = alt_seq[comb.prev_offset+comb.alt_offset:variant.offset+comb.alt_offset]
            inter_ref = ref_seq[comb.prev_offset:variant.offset]
            if not inter_alt == inter_ref:
                continue
            new_combinations.append(Combination(
                comb.code+[0],
                comb.alt_offset,
                comb.prev_offset))
            if comb.prev_offset > variant.offset:
                continue
            for code, seq in enumerate(variant.alt):
                var_alt = alt_seq[variant.offset+comb.alt_offset:variant.offset+comb.alt_offset+len(seq)]
                if var_alt == seq.lower():
                    new_combinations.append(Combination(
                        comb.code+[code+1],
                        comb.alt_offset+len(seq)-len(variant.ref),
                        variant.offset+len(variant.ref)))

        combinations = new_combinations
        # assert all(len(t[0]) == j+1 for t in combinations), (j, combinations)
    real = []
    for comb in combinations:
        if not len(alt_seq)-comb.alt_offset == len(ref_seq):
            continue
        alt_stub = alt_seq[comb.prev_offset+comb.alt_offset:len(ref_seq)+comb.alt_offset]
        if not alt_stub == ref_seq[comb.prev_offset:len(ref_seq)]:
            continue
        real.append(comb)

    if not len(real):
        logging.info(alt_seq)
        logging.info(ref_seq)
        logging.info(variants)
        logging.info("---->")
        logging.info("%s / %s", real, combinations)
        return ["No Match"]

    haplotypes = []
    for valid in real:
        codes = valid[0]
        haplotypes.extend(find_valid_haplotypes(variants, codes))
        # f = None
        # for code, variant in zip(codes, variants):
        #     f = variant.precence.get_samples(code, f)
        # haplotypes.extend(f)

    return np.sort(np.unique(haplotypes))
    # codes = real[0][0]
    # f = None
    # for code, variant in zip(codes, variants):
    #     f = variant.precence.get_samples(code, f)
    # return f


def find_haplotype(seq, refseq, vcf, start, end):
    refseq = refseq.lower()
    if seq == refseq:
        return ["REF"]
    # logging.info("Finding haplotype for %s->%s, (%s, %s)", refseq, seq, start, end)
    return traverse_variants(seq, refseq, vcf.get_variants_from(start, end))
    cur_offset = 0
    cur_ref_offset = 0
    haplotypes = []

    possible_vars = []
    for i, variant in enumerate(vcf.get_variants_from(start, end)):
        possible_vars.append(variant)
        new_offset = variant.offset
        l = new_offset - cur_ref_offset
        if seq[cur_offset:cur_offset+l] != refseq[cur_ref_offset:new_offset]:
            logging.error("%s != %s", seq[cur_offset:cur_offset+l], refseq[cur_ref_offset:new_offset])
            return ["ERROR"]
        cur_offset += l
        cur_ref_offset = new_offset
        if seq[cur_offset:cur_offset+len(variant.alt)] == variant.alt.lower():
            if (not seq[cur_offset:cur_offset+len(variant.ref)] == variant.ref.lower()) or len(variant.alt) > len(variant.ref):
                haplotypes.append(1)
                cur_offset += len(variant.alt)
                cur_ref_offset += len(variant.ref)
                continue
        # assert seq[cur_offset:cur_offset+len(variant.ref)] == variant.ref
        haplotypes.append(0)
    logging.info(possible_vars)
    logging.info([var for var, h in zip(possible_vars, haplotypes) if h])
    return haplotypes


class Main:

    def __init__(self, indexed_interval, graph, seq_graph, fasta, vcf):
        self.indexed_interval = indexed_interval
        self.graph = graph
        self.seq_graph = seq_graph
        self.fasta = fasta
        self.vcf = vcf
        self.counter = 0
        self.prev_end = -1

    def run_peak_set(self, peaks_dict):
        all_name_tuples = [(key, self.get_sequence_pair(read))
                           for key, reads in peaks_dict.items() for read in reads]
        names, tuples = zip(*sorted(all_name_tuples, key=lambda x: x[1][2][0]))
        alts, refs, intervals = zip(*tuples)
        params = zip(alts, refs, self.vcf.get_variants_from_intervals(intervals))
        haplotypes = (traverse_variants(*param) for param in params)
        haplotypes = (h if h is not None else list(range(1024)) for h in haplotypes)
        result_dict = {name: [] for name in peaks_dict}
        for name, result in zip(names, haplotypes):
            result_dict[name].append(result)
        return result_dict

    def run_peaks(self, peaks):
        alts, refs, intervals = zip(*(self.get_sequence_pair(peak) for peak in peaks))
        haplotypes = [traverse_variants(alt.lower(), ref.lower(), variants)
                      for alt, ref, variants in
                      zip(alts, refs, self.vcf.get_variants_from_intervals(intervals))]
        return haplotypes

    def run_peak(self, peak):
        seq, ref_seq, interval = self.get_sequence_pair(peak)
        if interval[0] < self.prev_end:
            return []
        self.prev_end = interval[-1]
        haplo = find_haplotype(seq, ref_seq, self.vcf, interval[0], interval[1])
        if haplo:
            self.counter += 1
        return haplo

    def extend_peak_to_reference(self, peak):
        start_node = peak.region_paths[0]
        end_node = peak.region_paths[-1]
        start_offset = peak.start_position.offset
        end_offset = peak.end_position.offset
        rps = peak.region_paths
        if start_node not in self.indexed_interval.nodes_in_interval():
            start_node = min(
                -node for node in self.graph.reverse_adj_list[-start_node]
                if -node in self.indexed_interval.nodes_in_interval())
            start_offset = self.graph.node_size(start_node)-1
            rps = [start_node]+rps
        if end_node not in self.indexed_interval.nodes_in_interval():
            end_node = max(node for node in self.graph.adj_list[end_node]
                           if node in self.indexed_interval.nodes_in_interval())
            end_offset = 1
            rps = rps + [end_node]
        return obg.Interval(start_offset, end_offset, rps)

    def get_sequence_pair(self, peak):
        peak.graph = self.graph
        peak = self.extend_peak_to_reference(peak)
        seq = self.seq_graph.get_interval_sequence(peak)
        assert len(seq) == peak.length(), (len(seq), peak.length())
        interval = (self.indexed_interval.get_offset_at_node(peak.region_paths[0]) + peak.start_position.offset,
                    self.indexed_interval.get_offset_at_node(peak.region_paths[-1]) + peak.end_position.offset)
        ref_seq = self.fasta[int(interval[0]):int(interval[1])].seq
        assert len(ref_seq) == interval[1]-interval[0]
        return seq.lower(), ref_seq.lower(), interval

    @classmethod
    def from_name(cls, name, fasta_file_name, chrom):
        indexed_interval = obg.NumpyIndexedInterval.from_file(
            name+"_linear_pathv2.interval")
        graph = obg.Graph.from_file(name+".nobg")
        seq_graph = obg.SequenceGraph.from_file(name+".nobg.sequences")
        fasta = Fasta(fasta_file_name)[chrom]
        vcf = VCF(name + "_variants.vcf")
        return cls(indexed_interval, graph, seq_graph, fasta, vcf)
    # var_list.create_lookup(name+"_variants.vcf")
    # return var_list


class DummyVCF:

    @staticmethod
    def get_variants_from(start):
        return [Variant(2, "A", "T"), Variant(4, "G", ""), Variant(5, "G", "C")]


def test():
    refseq = "AAAGGGTTT"
    seq = "AATGGCTTT"
    assert find_haplotype(seq, refseq, DummyVCF, 0) == [1, 0, 1]
    print("SUCCESS")

def test2():
    refseq = "AAAGGGTTT"
    seq = "AAAGGCTTT"
    assert find_haplotype(seq, refseq, DummyVCF, 0) == [0, 0, 1]
    print("SUCCESS")

def test3():
    refseq = "AAAGGGTTT"
    seq = "AAAGCTTT"
    assert find_haplotype(seq, refseq, DummyVCF, 0) == [0, 1, 1]
    print("SUCCESS")


def test4():
    vcf = "/home/knut/Documents/phd/data/graphs/1_variants.vcf"
    vcf = VCF(vcf)
    for i, var in enumerate(vcf.get_variants_from(70)):
        print(var)
        if i > 10:
            break


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    h = "./.:.:.	1|1:40:8	1|1:40:11	0|0:40:9	./.:.:.	./.:.:.	./.:.:.	0|0:40:8	0|0:40:9	./.:.:.	0|0:40:12	0|0:40:16	0|0:40:40	0|0:40:35	0|0:40:21	0|0:40:13	0|0:40:44	0|0:40:15	0|0:40:86	0|0:40:69	./.:.:.	0|0:40:14	0|0:40:59	0|0:40:51"
    # precence = VariantPrecence.from_line(h)
    # print(precence._precence)
    # print(precence.get_samples(1, [0, 1, 3]))
    alt = "ccttctttttttg"
    ref = "ccttctttttttg"
    variants = [FullVariant(offset=0, ref='ctt', alt=['ttt', 'c', 'ctttt'], precence=None),
                FullVariant(offset=2, ref='t', alt=['tc'], precence=None),
                FullVariant(offset=5, ref='t', alt=['c'], precence=None),
                FullVariant(offset=6, ref='t', alt=['a'], precence=None),
                FullVariant(offset=7, ref='t', alt=['c', 'tc'], precence=None)]
    # alt = "atgcctttattatccttcacgttgaccccacatgccccttttttttttttgg"
    # ref = "atgcctttattatccttcacgttgaccccacatgcccctgttttttttttttg"
    # variants = [FullVariant(offset=2, ref='g', alt=['a'], precence=None),
    #             FullVariant(offset=18, ref='a', alt=['t'], precence=None),
    #             FullVariant(offset=30, ref='c', alt=['t'], precence=None),
    #             FullVariant(offset=38, ref='tgtt', alt=['ttt', 'tttt', 'tgt', 'tg'], precence=None),
    #             FullVariant(offset=44, ref='t', alt=['c'], precence=None),
    #             FullVariant(offset=45, ref='t', alt=['c'], precence=None)]

    print(traverse_variants(alt, ref, variants))


    #alt = "aaaaaataagacgt"
    #ref = "ataaataagacgt"
    #variants = [Variant(offset=1, ref='T', alt=['A']), Variant(offset=8, ref='GACGTACCCTCA', alt=['G', 'GAGGTACCCTCA'])]

    # alt = "cccttctttttttg"
    # ref = "cttttttttttttg"
    # variants = [Variant(offset=0, ref='CTT', alt=['TTT', 'C', 'CTTTT', 'CTTC']), Variant(offset=5, ref='T', alt=['C']), Variant(offset=6, ref='T', alt=['A']), Variant(offset=7, ref='T', alt=['C', 'TC'])]

    # alt = "ggaaataaaaaa"
    # ref = "ggaaataaaaa"
    # variants = [Variant(offset=5, ref='T', alt=['C', 'TA']), Variant(offset=7, ref='A', alt=['C']), Variant(offset=8, ref='A', alt=['T', 'G']), Variant(offset=11, ref='T', alt=['A'])]

    # alt = "agtttcactggg"
    # ref = "agtttcagtagg"
    # variants = [Variant(offset=7, ref='GTA', alt=['CTA', 'ATA', 'GT', 'GTG'])]
    # [Variant(offset=7, ref='G', alt=['C', 'A']), Variant(offset=8, ref='TA', alt=['T']), Variant(offset=9, ref='A', alt=['G'])]

#     alt = "accttatagaaa"
#     ref = "accttataagaaa"
#     variants = [Variant(offset=6, ref='TA', alt=['T'])]

    # alt = "agtttcactggg"
    # ref = "agtttcagtagg"
    # variants = [Variant(offset=7, ref='G', alt=['C', 'A']), Variant(offset=9, ref='A', alt=['', 'G'])]
    # print(traverse_variants(alt, ref, variants))

# ct at 0x7fba28385c18>),
# t at 0x7fb8521fe9e8>),
# t at 0x7fba2e1d35f8>), 
# recence object at 0x7fba2e1d3860>), 
# t at 0x7fba28385278>), 
# ect at 0x7fba283852b0>), 
# t at 0x7fba280af4e0>), 
# t at 0x7fba28385940>), 
# t at 0x7fba28385080>)]



# 2018-09-03 19:19:20,165, INFO: atgcctttattatccttcacgttgaccccacatgccccttttttttttttgg
# 2018-09-03 19:19:20,165, INFO: atgcctttattatccttcacgttgaccccacatgcccctgttttttttttttg
# 2018-09-03 19:19:20,165, INFO: 
# variants = [FullVariant(offset=2, ref='g', alt=['a'], precence=None), 
#             FullVariant(offset=18, ref='a', alt=['t'], precence=None),
#             FullVariant(offset=30, ref='c', alt=['t'], precence=None),
#             FullVariant(offset=38, ref='tgtt', alt=['ttt', 'tttt', 'tgt', 'tg'], precence=None),
#             FullVariant(offset=44, ref='t', alt=['c'], precence=None),
#             FullVariant(offset=45, ref='t', alt=['c'], precence=None)]

# 2018-09-03 19:19:20,165, INFO: [] / [([0, 0, 0, 1, 0, 0], -1, 46), ([0, 0, 0, 2, 0, 0], 0, 46)]


# 2018-09-03 21:09:38,315, INFO: [] / [([0, 0, 0, 0, 0, 0], 0, 43), ([0, 0, 0, 1, 0, 0], 1, 43)]
# 2018-09-03 21:09:38,327, INFO: ttttttgatagattatatcaaatccatggatactttctatatttggaaagt
# 2018-09-03 21:09:38,328, INFO: tttttttgatagattatatcaaatccatggatactttctatatttggaaagt
# 2018-09-03 21:09:38,328, INFO: 
# [FullVariant(offset=0, ref='t', alt=['ta'], precence=None),
#  FullVariant(offset=6, ref='tg', alt=['gg', 't', 'tt'], precence=None),
#  FullVariant(offset=26, ref='a', alt=['g'], precence=None),
#  FullVariant(offset=37, ref='c', alt=['ct'], precence=None)]
# 
# r = "taaaaacaaagaaagtcaactaccctattc"
# 
# a = "taaaaaaacaaagaaagtcaactaccctattc"
# r = "tggaaacaaagaaagtcaactaccctattc"
# 
# 
# [FullVariant(offset=1, ref='g', alt=['a'], precence=None),
#  FullVariant(offset=2, ref='g', alt=['a', 'gaa'], precence=None),
#  FullVariant(offset=6, ref='c', alt=['t'], precence=None),
#  FullVariant(offset=14, ref='g', alt=['c'], precence=None),
#  FullVariant(offset=19, ref='c', alt=['t'], precence=None),
#  FullVariant(offset=24, ref='c', alt=['a'], precence=None)]
# 2018-09-03 22:26:09,814, INFO: ---->
# 2018-09-03 22:26:09,814, INFO: [] / []


# 
# 
# 
# 
# 
# 
# 
# 
# 
# 
# code=[1, 0, 0, 1, 1, 1], alt_offset=17, prev_offset=29)]
#  alt = "ccttctttttttg"
#  ref = "ccttctttttttg"
#  [FullVariant(offset=0, ref='ctt', alt=['ttt', 'c', 'ctttt'], precence=None),
#   FullVariant(offset=2, ref='t', alt=['tc'], precence=None),
#   FullVariant(offset=5, ref='t', alt=['c'], precence=None),
#   FullVariant(offset=6, ref='t', alt=['a'], precence=None),
#   FullVariant(offset=7, ref='t', alt=['c', 'tc'], precence=None)]


# alt = "taaaaagagagagaggtatagaggaaaaagagaaagagataaagaaagctat"
# ref = "taaaaaggagagagaggtatagaggaaaaagagaaagagataaagaaagctat"
# variants = [
#     FullVariant(offset=1, ref='t', alt=['a'], precence=P(64)), 
#     FullVariant(offset=6, ref='ag', alt=['gg', 'a'], precence=P(4)), 
#     FullVariant(offset=9, ref='g', alt=['a'], precence=P(362)), 
#     FullVariant(offset=37, ref='a', alt=['g'], precence=P(64)), 
#     FullVariant(offset=50, ref='t', alt=['c'], precence=P(26)), 
#     FullVariant(offset=51, ref='a', alt=['t'], precence=P(2))]
# 
# 
# ref = "tcacaacttatgcctattgaattatttagtgtcccgtcgtcggaatcagt"
# alt = "tcacaacttatgcctattgaattatttagtgtcccgtcgtcggaatcagtt"
# ref = "tcacaacttatgcctattgaattatttagtgtcccgtcgtcggaatcagt"
# variants = [FullVariant(offset=18, ref='g', alt=['a'], precence=None),
#             FullVariant(offset=19, ref='a', alt=['t'], precence=None),
#             FullVariant(offset=32, ref='c', alt=['g'], precence=None),
#             FullVariant(offset=40, ref='c', alt=['a'], precence=None)]
# 
# 
# ref = "ataaaaaaaggaagagattgaattgtgtgaaacctggaggaagcggagccagt"
# alt = "aaaaaaaggaagagattgaattgtgtgaaacctggaggaagcggagccagt"
# ref = "ataaaaaaaggaagagattgaattgtgtgaaacctggaggaagcggagccagt"
# variants = [FullVariant(offset=0, ref='at', alt=['a'], precence=None),
#             FullVariant(offset=1, ref='ta', alt=['t'], precence=None),
#             FullVariant(offset=2, ref='a', alt=['c'], precence=None),
#             FullVariant(offset=3, ref='a', alt=['g'], precence=None),
#             FullVariant(offset=29, ref='a', alt=['ag'], precence=None),
#             FullVariant(offset=30, ref='a', alt=['at'], precence=None),
#             FullVariant(offset=32, ref='c', alt=['a'], precence=None)]



import logging
import numpy as np


class SparseValues:
    def __init__(self, indices, values, sanitize=False):
        self.indices = np.asanyarray(indices)
        self.values = np.asanyarray(values)
        if sanitize:
            self._sanitize()
        self.track_size = None

    def _sanitize(self):
        unique_mask = np.r_[self.indices[1:] != self.indices[:-1], True]
        self.indices = self.indices[unique_mask]
        self.values = self.values[unique_mask]

        values_mask = np.r_[True, self.values[1:] != self.values[:-1]]
        self.indices = self.indices[values_mask]
        self.values = self.values[values_mask]

    @classmethod
    def from_sparse_files(cls, file_base_name):
        indices = np.load(file_base_name + "_indexes.npy")
        values = np.load(file_base_name + "_values.npy")
        size = indices[-1]
        obj = cls(indices[:-1], values)
        obj.track_size = size
        return obj

    def to_sparse_files(self, file_base_name):
        np.save(file_base_name + "_indexes.npy",
                np.r_[self.indices, self.track_size])
        np.save(file_base_name + "_values.npy", self.values)
        logging.info("Wrote to %s_indexes/values.npy" % file_base_name)

    def __repr__(self):
        return "SV(%s, %s)" % (self.indices, self.values)

    def __eq__(self, other):
        if not np.all(self.indices == other.indices):
            return False

        if not np.allclose(self.values, other.values):
            return False

        if self.track_size != other.track_size:
            return False

        return True

    def to_bed_graph(self, filename):
        logging.warning("Not writing to %s", filename)

    def to_bed_file(self, filename):
        logging.warning("Not writing to %s", filename)

    def threshold_copy(self, cutoff):
        values = self.values >= cutoff
        new = SparseValues(self.indices, values, sanitize=True)
        new.track_size = self.track_size
        return new

    def to_dense_pileup(self, size):
        if self.values.dtype == np.bool:
            values = self.values.astype("int")
        else:
            values = self.values
        diffs = np.ediff1d(values, to_begin=values[0])
        pileup = np.zeros(size+1, dtype=values.dtype)
        indices = self.indices[:diffs.size]
        pileup[indices] = diffs
        pileup = np.cumsum(pileup[:-1])
        if self.values.dtype == np.bool:
            pileup = pileup.astype("bool")
        return pileup


    @classmethod
    def from_dense_pileup(cls, pileup):
        changes = pileup[1:] != pileup[:-1]
        indices = np.r_[0, np.flatnonzero(changes)+1]
        values = pileup[indices]
        obj = cls(indices, values)
        obj.track_size = pileup.size
        return obj


class SparseDiffs:
    def __init__(self, indices, diffs, sanitize=False):
        self._indices = np.asanyarray(indices)
        self._diffs = np.asanyarray(diffs)
        if sanitize:
            self._sanitize()

    def to_dense_pileup(self, size):
        valued = self.get_sparse_values()
        return valued.to_dense_pileup(size)
        # pileup = np.zeros(size+1, dtype=self._diffs.dtype)
        # pileup[self._indices] = self._diffs
        # return np.cumsum(pileup[:-1])

    def to_sparse_files(self, file_base_name):
        np.save(file_base_name + "_indexes.npy",
                self._indices)
        np.save(file_base_name + "_values.npy", self._diffs)
        logging.info("Wrote to %s and %s" % (file_base_name + "_indexes.npy",
                                             file_base_name + "_values.npy"))

    @classmethod
    def from_sparse_files(cls, file_base_name):
        indices = np.load(file_base_name + "_indexes.npy")
        values = np.load(file_base_name + "_values.npy")
        size = indices[-1] + 10  # hack
        obj = cls(indices[:-1], values)
        obj.track_size = size
        return obj.get_sparse_values()

    def to_bed_graph(self, filename):
        logging.warning("Not writing to %s", filename)

    def __repr__(self):
        return "SD(%s, %s)" % (self._indices, self._diffs)

    def clean(self):
        sparse_values = self.get_sparse_values()
        self._indices = sparse_values.indices
        self._diffs = np.ediff1d(sparse_values.values,
                                 to_begin=sparse_values.values[0])

    def clip_min(self, min_value):
        values = np.cumsum(self._diffs)
        np.clip(values, min_value, None, values)
        self._diffs[1:] = np.diff(values)
        self._diffs[0] = values[0]
        self._sanitize()

    def get_sparse_values(self):
        args = np.argsort(self._indices, kind="mergesort")
        diffs = self._diffs[args]
        values = np.cumsum(diffs)
        return SparseValues(self._indices[args], values, sanitize=True)


    def maximum(self, other):
        all_indices = np.r_[self._indices, other._indices]
        sorted_args = np.argsort(all_indices, kind="mergesort")
        new_indexes = all_indices[sorted_args]
        unique_mask = np.ediff1d(new_indexes, to_end=1) != 0
        new_indexes = new_indexes[unique_mask]
        diffs = np.zeros((2, all_indices.size))
        my_args = sorted_args < self._indices.size
        other_args = ~my_args
        diffs[0][my_args] = self._diffs
        diffs[1][other_args] = other._diffs
        values = np.cumsum(diffs, axis=1)
        max_values = np.max(values, axis=0)
        max_values = max_values[unique_mask]
        new_diffs = np.ediff1d(
            max_values, to_begin=max_values[0])
        return SparseDiffs(new_indexes, new_diffs, True)

    def _sanitize(self):
        # Remove duplicated values
        changes = self._diffs != 0
        self._indices = self._indices[changes]
        self._diffs = self._diffs[changes]

    def __eq__(self, other):
        return np.all(self._indices == other._indices)\
            and np.all(self._diffs == other._diffs)

    @classmethod
    def from_starts_and_ends(cls, starts_ends):
        indices = starts_ends.ravel(order="C")
        args = np.argsort(indices)
        indices = indices[args]
        diffs = np.ones(indices.size)
        diffs[args >= starts_ends.shape[1]] *= -1
        return cls(indices, diffs)

    @classmethod
    def from_pileup(cls, pileup, node_indexes):
        indices = np.r_[pileup.starts, pileup.ends, node_indexes]

        diffs = np.r_[np.ones(len(pileup.starts)),
                      -1*np.ones(len(pileup.ends)),
                      pileup.node_starts]
        args = np.argsort(indices)
        indices = indices[args]
        diffs = diffs[args]
        return cls(indices, diffs)

    @classmethod
    def from_dense_pileup(cls, pileup):
        diffs = np.ediff1d(pileup, to_begin=1)
        changes = np.flatnonzero(diffs)
        diffs = diffs[changes]
        diffs[0] = pileup[0]
        return cls(changes, diffs)

    def __itruediv__(self, scalar):
        self._diffs /= scalar
        return self

    def __imul__(self, scalar):
        self._diffs *= scalar
        return self

    def apply_binary_func(self, func, other, return_values=False):
        all_indices = np.r_[self._indices, other._indices]
        sorted_args = np.argsort(all_indices, kind="mergesort")
        new_indexes = all_indices[sorted_args]
        unique_mask = np.ediff1d(new_indexes, to_end=1) != 0
        new_indexes = new_indexes[unique_mask]
        diffs = np.zeros((2, all_indices.size))
        my_args = sorted_args < self._indices.size
        other_args = ~my_args
        diffs[0][my_args] = self._diffs
        diffs[1][other_args] = other._diffs
        values = np.cumsum(diffs, axis=1)
        ret = func(values[0], values[1])[unique_mask]
        if return_values:
            return SparseValues(new_indexes, ret, sanitize=True)
        return SparseDiffs(
            new_indexes, np.ediff1d(ret, to_begin=ret[0]))



import pickle
import numpy as np
import os
from collections import Counter
from .densepileup import DensePileup
from scipy.stats import poisson
import logging
import math


class PValuesFinder:
    def __init__(self, sample_pileup, control_pileup):
        self.sample = sample_pileup
        self.control = control_pileup

    def get_p_values_pileup(self):
        p_values = poisson.logsf(
            self.sample.data._values,
            self.control.data._values)
        baseEtoTen = np.log(10)
        p_values_array = -p_values / baseEtoTen

        # Set p-value to 0 where sample is 0
        zero_indices = np.where(self.sample.data._values == 0)[0]
        p_values_array[zero_indices] = 0

        p_values_pileup = DensePileup(self.sample.graph)
        p_values_pileup.set_new_values(p_values_array)
        p_values_pileup.data._touched_nodes = self.sample.data._touched_nodes
        return p_values_pileup


class PToQValuesMapper:

    def __init__(self, counter):
        self._counter = counter

    @classmethod
    def __read_file(cls, file_name):
        indices = np.load(file_name + "_indexes.npy")
        values = np.load(file_name + "_values.npy")
        return indices, values

    @classmethod
    def from_p_values_dense_pileup(cls, p_values):
        logging.info("Creating mapping from p value dense pileup")
        values = p_values.data._values
        sorted_p_values = sorted(values, reverse=True)
        unique, counts = np.unique(sorted_p_values, return_counts=True)
        counter_dict = {unique[i]: counts[i]
                        for i in range(0, len(counts))}
        return cls(counter_dict)

    @classmethod
    def from_files(cls, base_file_name):
        search = base_file_name
        logging.info("Searching for files starting with %s" % search)
        files = (f for f in os.listdir()
                 if f.startswith(search) and "pvalues" in f and f.endswith("_indexes.npy"))
        counter = Counter()
        for filename in files:
            base_file_name = filename.replace("_indexes.npy", "")  # Get file name base
            logging.info("Reading p values from file %s" % base_file_name)
            indices, values = cls.__read_file(base_file_name)
            counts = np.diff(indices)
            for key, val in zip(values, counts):
                counter[key] += val
            #counter.update(dict(zip(values, counts)))
        return cls(counter)

    def get_p_to_q_values(self):
        logging.info("Mapping p to q values")
        p_to_q_values = {}
        rank = 1
        logN = np.log10(sum(self._counter.values()))
        pre_q = None
        i = 0
        for p_value in reversed(sorted(self._counter.keys())):
            if i % 5000 == 0:
                logging.info("Done mapping p value %d/%d" % (i, len(self._counter.keys())))
            i += 1

            value_count = self._counter[p_value]
            q_value = p_value + (np.log10(rank) - logN)
            if rank == 1:
                q_value = max(0.0, q_value)
            else:
                q_value = max(0.0, min(pre_q, q_value))
            p_to_q_values["%.7f" % p_value] = q_value
            pre_q = q_value
            rank += value_count

        self.p_to_q_values = p_to_q_values
        logging.info("Done creating mapping")
        return p_to_q_values

    def to_file(self, base_name):
        with open(base_name + 'p2q.pkl', 'wb') as f:
            pickle.dump(self._p_to_q_values, f, pickle.HIGHEST_PROTOCOL)


class QValuesFinder:
    def __init__(self, p_values_pileup, p_to_q_values):
        assert isinstance(p_to_q_values, dict)
        self.p_values = p_values_pileup
        self.p_to_q_values = p_to_q_values

    def get_q_values(self):
        q_values_pileup = DensePileup(self.p_values.graph)
        new_values = self.get_q_array_from_p_array(
                        self.p_values.data._values)
        q_values_pileup.set_new_values(new_values)
        q_values_pileup.data._touched_nodes = self.p_values.data._touched_nodes
        return q_values_pileup

    def get_q_array_from_p_array(self, p_values):
        assert isinstance(p_values, np.ndarray)

        def translation(x):
            if abs(x) < 1e-9:
                return 0
            if math.isnan(x):
                return 0
            x = "%.7f" % x
            if x not in self.p_to_q_values:
                print(self.p_to_q_values)
                logging.error("P value not found in mapping dict. Could be due to rounding errors.")
            return self.p_to_q_values[x]

        trans = np.vectorize(translation, otypes=[np.float])
        new_values = trans(p_values)
        return new_values



#from gendatafetcher.sequences import get_sequence_ucsc
from Bio import Seq, SeqIO, SeqRecord
import logging
from pyfaidx import Fasta
import json
import numpy as np


class NonGraphPeak:
    def __init__(self, chromosome, start, end, score=None, unique_id=None):
        self.chromosome = chromosome
        self.start = start
        self.end = end
        self.score = score
        self.sequence = None
        self.unique_id = None

    def set_sequence_using_fasta_index(self, fasta_index):
        self.sequence = str(fasta_index[self.chromosome][self.start:self.end])

    def __str__(self):
        return "Peak(%s, %d, %d, score=%s)" % (self.chromosome,
                                                 self.start,
                                                 self.end,
                                                 self.score)

    def __eq__(self, other):
        if self.chromosome != other.chromosome:
            return False
        if self.start != other.start:
            return False
        if self.end != other.end:
            return False

        return True

    def to_bed_line(self):
        name = self.unique_id
        if name is None:
            name = "."

        return "%s\t%d\t%d\t%s\t.\t.\t.\t.\t%.2f\n" % (self.chromosome, self.start, self.end, name, self.score)

    def to_file_line(self):
        object = {
                "chromosome": self.chromosome,
                "start": int(self.start),
                "end": int(self.end),
                "score": float(self.score)
                  }
        try:
            d = json.dumps(object)
        except:
            for k, v in object.items():
                print(k, v, type(v))
            raise
        return d

    @classmethod
    def from_file_line(cls, line):
        object = json.loads(line)
        return cls(object["chromosome"], object["start"], object["end"],
                   score=object["score"])

    def __repr__(self):
        return self.__str__()


class NonGraphPeakCollection(object):
    def __init__(self, peaks):
        self.peaks = peaks

    @classmethod
    def from_bed_file(cls, file_name, cut_around_summit=None):
        peaks = []
        f = open(file_name)
        for line in f:
            peak = line.split()
            chrom = peak[0]
            start = int(peak[1])
            end = int(peak[2])

            if len(peak) >= 7:
                score = float(peak[8])  # q value
            else:
                score = -1


            if cut_around_summit is not None:
                summit = int(peak[9])
                old_start = start
                start = max(start, start + summit - cut_around_summit)
                end = min(end, old_start + summit + cut_around_summit)
                assert start + cut_around_summit * 2 >= end, "End %d is > start + 2x summit" % (end)
                assert end > start

            peaks.append(NonGraphPeak(chrom,
                                      start,
                                      end,
                                      score))
        f.close()
        return cls(peaks)

    def to_bed_file(self, file_name):
        f = open(file_name, "w")
        for peak in self.peaks:
            f.writelines([peak.to_bed_line()])
        f.close()
        logging.info("Wrote to bed-file %s" % file_name)

    def get_graph_peak_collection(self, graph, linear_interval_through_graph, graph_region=None):
        from .peakcollection import PeakCollection
        return PeakCollection.create_from_nongraph_peak_collection(graph, self,
                                                                   linear_interval_through_graph,
                                                                   graph_region)


    def filter_peaks_outside_region(self, chromosome, start, end):
        new_peaks = []
        for peak in self.peaks:
            if peak.chromosome != chromosome:
                continue
            if peak.start < start or peak.end > end:
                continue
            new_peaks.append(peak)
        self.peaks = new_peaks

    def _sort_on_score_descending(self):
        self.peaks.sort(key=lambda peak: peak.score, reverse=True)

    def to_fasta(self, file_name):
        """
        lines = (SeqRecord.SeqRecord(Seq.Seq(peak.sequence),
                                     id=str(i), description=peak.to_file_line())
                 for i, peak in enumerate(self.peaks))

        SeqIO.write(lines, file_name, "fasta")
        """
        f = open(file_name, "w")
        for i, peak in enumerate(self.peaks):
            if peak.unique_id is None:
                id = i
            else:
                id = peak.unique_id
            f.writelines([">%d %s\n" % (id, peak.to_file_line())])
            f.writelines(["%s\n" % peak.sequence])
        f.close()

    @classmethod
    def from_fasta(cls, file_name):
        f = open(file_name)
        peaks = []
        while True:
            header = f.readline()
            sequence = f.readline()
            if not sequence:
                break

            header = header.split(maxsplit=1)
            id = header[0].replace(">", "")
            interval_json = header[1]
            peak = NonGraphPeak.from_file_line(interval_json)
            peak.unique_id = id
            assert peak.unique_id is not None
            peak.sequence = sequence
            peaks.append(peak)

        avg_peak_size = np.mean([p.end - p.start for p in peaks])
        logging.info("Avg peak size: %.2f" % avg_peak_size)

        return cls(peaks)

    def set_peak_sequences_using_fasta(self, fasta_file_location="grch38.fasta"):
        logging.info("Setting peak sequences using fasta index")
        genome = Fasta(fasta_file_location)
        i = 0
        for peak in self.peaks:
            if i % 10000 == 0:
                logging.info("%d/%d peaks processed" % (i, len(self.peaks)))
            i += 1

            peak.set_sequence_using_fasta_index(genome)

    def save_to_sorted_fasta(self, file_name):
        self._sort_on_score_descending()
        self.to_fasta(file_name)



import matplotlib
matplotlib.use("agg")
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.patches import Rectangle, Polygon, FancyBboxPatch

from matplotlib.collections import PatchCollection


def square_venn(ax, gpc, macs, shared, m_gpc_u, m_gpc_s, m_macs_u, m_macs_s):
    col1 = (0.8, 0.3, 0.3)
    col2 = (0.3, 0.8, 0.3)
    total = float(gpc+macs-shared)
    r_gpc = (0, gpc, 1, col1)
    r_macs = (gpc-shared, macs, 1, col2)
    r_m_gpc_u = (0, gpc-shared, m_gpc_u/float(gpc-shared), col1)
    r_m_macs_u = (gpc, total-gpc, m_macs_u/float(macs-shared), col2)
    coords = [r_gpc, r_macs]  # , r_m_macs_u]  # , r_m_gpc_s, r_m_macs_s]
    coords = [(x/total, w/total, h, c) for x, w, h, c in coords]
    print(coords)
    pad = 0.01
    recs = [FancyBboxPatch((x+pad, 0+pad), w-2*pad, h-2*pad, alpha=0.3, facecolor=c,
                           linestyle="dashed", boxstyle="round,pad=0.01")
            for x, w, h, c in coords]
    # recs = [FancyBboxPatch((x, 0), w, h, alpha=0.3, facecolor=c,
    #                        linestyle="dashed", boxstyle="round, pad=0.0")
    #         for x, w, h, c in coords]

    p_gpc_u = m_gpc_u/float(gpc-shared)
    p_gpc_s = m_gpc_s/float(shared)
    p_macs_u = m_macs_u/float(macs-shared)
    p_macs_s = m_macs_s/float(shared)
    xy_gpc_s = [(0, 0),
                (gpc/total, 0),
                (gpc/total, p_gpc_s),
                ((gpc-shared)/total, p_gpc_s),
                ((gpc-shared)/total, p_gpc_u),
                (0, p_gpc_u)]
    xy_macs_s = [(1, 0),
                 (1-macs/total, 0),
                 (1-macs/total, p_macs_s),
                 (gpc/total, p_macs_s),
                 (gpc/total, p_macs_u),
                 (1, p_macs_u)]
    polygs = []
    polygs.append(
        Polygon(xy=xy_gpc_s, facecolor=col1, alpha=0.7))
    polygs.append(
        Polygon(xy=xy_macs_s, facecolor=col2, alpha=0.7))

    # r_m_gpc_s = ((gpc-shared, 0), (gpc, u_gpc_s/float(shared)), 
    # r_m_gpc_s = (gpc-shared, shared, m_gpc_s/float(shared), "g")
    # r_m_macs_s = (gpc-shared, shared, m_macs_s/float(shared), "r")
    # collection = PatchCollection(recs, alpha=0.3)
    # collection.set_array(np.array([0, 0.25, 0.5, 0.75, 0.5, 0.75]))
    # collection.set_array(np.linspace(0, 1, len(recs)))
    for rec in recs + polygs:
        ax.add_patch(rec)

    return (gpc-shared)/(2*total), (gpc-0.5*shared)/total, (gpc+(macs-shared)/2.)/total
    # ax.add_collection(collection)

def column_venn(ax, gpc, macs, shared, m_gpc_u, m_gpc_s, m_macs_u, m_macs_s, both_motif):
    total = gpc+macs-shared
    xs = [0, (gpc-shared)/total, (gpc+macs-2*shared)/total]
    widths = [(gpc-shared)/total, (macs-shared)/total, shared/total]
    print(xs, widths)
    heights = [m_gpc_u/(gpc-shared), m_macs_u/(macs-shared), both_motif/shared]
    colors = ["b", "#db9200", "g"]
    big_rectangles = [Rectangle((x, 0), w, 1, facecolor=c, alpha=0.3)
                      for x, w, c in zip(xs, widths, colors)]
    small_rectangles = [Rectangle((x,0), w, h, facecolor=c, alpha=0.3)
                        for x, w, h, c in zip(xs, widths, heights, colors)]
    last_h = (m_gpc_s+m_macs_s-2*both_motif)/shared
    ws = [widths[2]*m_gpc_s/(m_gpc_s+m_macs_s), widths[2]*m_macs_s/(m_gpc_s+m_macs_s)]
    last_recs = [Rectangle((xs[2], heights[2]), ws[0], last_h,
                           facecolor="b", alpha=0.6),
                 Rectangle((xs[2]+ws[0], heights[2]), ws[1], last_h,
                           facecolor="#e0b721", alpha=0.6)]

    for rec in big_rectangles+small_rectangles+last_recs:
        ax.add_patch(rec)
    return [x+w*0.5 for x, w in zip(xs, widths)]


def from_csv(filename):
    lines = [line.split() for line in open(filename)]
    counts = [float(n) for n in lines[1][1:]]
    print(counts)
    names = lines[0][1:]
    d = dict(zip(names, counts))
    numbers = counts[:3]+[
        d["MOTIF_UNIQUE_GPC"],
        d["MOTIF_SHARED_GPC"],
        d["MOTIF_UNIQUE_MACS"],
        d["MOTIF_SHARED_MACS"],
        d["MOTIF_BOTH"]]
    return numbers


def save_venn(numbers, out_name):
    fig, ax = plt.subplots(1)
    xs = column_venn(ax, *numbers)
    #plt.text(xs[0], 0.8, "GPC", ha="center")
    #plt.text(xs[1], 0.8, "MACS", ha="center")
    #plt.text(xs[2], 0.8, "SHARED", ha="center")
    try:
        # Try hacky way to get title
        title = out_name.split("/")[-1].split(".")[0].split("_")[1]
    except IndexError:
        title = "Untitled %s" % out_name

    plt.title(title)
    plt.savefig(out_name+"_venn.pdf", bbox_inches='tight')


def save_venn_from_csv(filename, out_name):
    numbers = from_csv(filename)
    save_venn(numbers, out_name)

def save_venn_from_csvs(filenames, out_name):
    #fig, axes = plt.subplots(2,3)
    #axes = [ax for row in axes for ax in row]
    numbers_list = [from_csv(filename) for filename in filenames]
    #for numbers, ax in zip(numbers_list, axes):
    #    column_venn(ax, *numbers)
    total = [sum(elems) for elems in zip(*numbers_list)]
    #print(total)
    #column_venn(axes[-1], *total)
    font = {'family' : 'DejaVu Sans',
        'weight' : 'normal',
        'size'   : 30}
    import matplotlib
    matplotlib.rc('font', **font)
    plt.figure(figsize=(15, 15.0), dpi=300)
    ax = plt.gca()
    column_venn(ax, *total)
    plt.show()
    plt.savefig(out_name)
    # plt.text(xs[0], 0.8, "GPC", ha="center")
    # plt.text(xs[1], 0.8, "MACS", ha="center")
    # plt.text(xs[2], 0.8, "SHARED", ha="center")
    # plt.savefig(out_name+"_venn.pdf", bbox_inches='tight')
    # numbers = from_csv(filename)
    # save_venn(numbers, out_name)

if __name__ == "__main__":
    import sys
    filenames = sys.argv[1]
    save_venn_from_csvs(filenames.split(","), sys.argv[2])



import sys

narrowpeak_file = open(sys.argv[1])
n_bp_each_side = int(sys.argv[2])
out_file = open(sys.argv[3], "w")

for line in narrowpeak_file:
    if line.startswith("track"):
        out_file.writelines([line])
        continue
    l = line.split()
    chrom = l[0]
    start = int(l[1])
    end = int(l[2])
    peak = int(l[9])

    new_start = max(start, start + peak - n_bp_each_side)
    new_end = min(end, start + peak + n_bp_each_side)

    assert new_end - new_start <= 2*n_bp_each_side + 1

    out_file.writelines(["chr%s\t%d\t%d\t%s\t%s\t.\t%s\t%s\t%s\t%s\n" % (chrom, new_start, new_end, l[3], l[4], l[6], l[7], l[8], (start + peak) - new_start)])


out_file.close()
narrowpeak_file.close()




import logging
import pickle
from collections import deque

from offsetbasedgraph.graphtraverser import GraphTravserserBetweenNodes
import offsetbasedgraph as obg

from pyvg import Snarls


class SnarlGraph(obg.GraphWithReversals):

    def __init__(self, blocks, edges, id=None, parent=None,
                 children=[], start_node=None, end_node=None):
        super(SnarlGraph, self).__init__(blocks, edges)

        for child in children:
            assert child.id != id, "Child id %d = self.id %d" % (child.id, id)
        self._start_node = start_node
        self._end_node = end_node
        self.parent = parent
        self.children = children
        self.id = id
        self.create_children()

        self._length = None

    def get_next_nodes(self, node_id):
        if node_id not in self.adj_list:
            return []
        return self.adj_list[node_id]

    def get_previous_nodes(self, node_id):
        return self.reverse_adj_list[node_id]

    def node_size(self, node_id):
        length = self.blocks[abs(node_id)].length()
        assert length > 0
        return length

    def get_all_leaf_node_ids(self):
        leaf_nodes = [node_id for node_id, obj in self.blocks.items()
                      if isinstance(obj, obg.Block)]
        subgraphs = [obj for obj in self.blocks.values()
                     if isinstance(obj, SnarlGraph)]

        for subgraph in subgraphs:
            leaf_nodes.extend(subgraph.get_all_leaf_node_ids())
        return leaf_nodes

    def create_children(self):

        n_childs = len(self.children)
        for i, child in enumerate(self.children):
            if self.id == "top_level" and (i % 20000 == 0 or i < 4):
                logging.info("Snarl %s of %s" % (i, n_childs))
            assert child.id != self.id, "Child ID %d equal as parent" % child.id
            child_blocks, child_graph = SnarlGraph.create_from_simple_snarl(
                child, self)

            if len(child_blocks) == 0:
                continue

            self.blocks[child.id] = child_graph

            for node_id in child_blocks.keys():
                if node_id == child.start or node_id == child.end:
                    continue

                del self.blocks[abs(node_id)]
                self._delete_edges_to_and_from_node(node_id)
                self._delete_edge(child.start, node_id)
                self._delete_edge(child.start, -node_id)
                self._delete_edge(-child.end, node_id)
                self._delete_edge(-child.end, -node_id)

            # add new edges
            self.adj_list[child.start].append(child.id)
            self.reverse_adj_list[-child.id] = [-child.start]

            self.adj_list[child.id] = [child.end]
            self.reverse_adj_list[-child.end].append(-child.id)

    def _remove_edges(self, node_id):
        if node_id in self.adj_list:
            edges = self.adj_list[node_id]
            for edge in edges:
                self.reverse_adj_list[-edge].remove(-node_id)
            del self.adj_list[node_id]

        if -node_id in self.reverse_adj_list:
            edges = self.reverse_adj_list[-node_id]
            for edge in edges:
                self.adj_list[-edge].remove(node_id)
            del self.reverse_adj_list[-node_id]

    def _delete_edges_to_and_from_node(self, node_id):
        self._remove_edges(node_id)
        self._remove_edges(-node_id)
        return

    def _delete_edge(self, from_node, to_node):
        if to_node in self.adj_list[from_node]:
            self.adj_list[from_node].remove(to_node)

        if -from_node in self.reverse_adj_list[-to_node]:
            self.reverse_adj_list[-to_node].remove(-from_node)

    def _sanitize_graph(self):
        # Check that all edges goes to blocks that exists
        for block, adj_list in self.adj_list.items():
            for edge in adj_list:
                assert edge in self.blocks, "Block from %d going to %d. %d does not exist." % (block, edge, edge)

    @classmethod
    def create_from_simple_snarl(cls, simple_snarl, parent_snarl_graph):
        traverser = GraphTravserserBetweenNodes(parent_snarl_graph)
        subgraph = traverser.get_snarl_subgraph(
            simple_snarl.start, simple_snarl.end, include_start_and_end=True)

        # Remove start and end blocks (keep edges)
        del subgraph.blocks[simple_snarl.start]
        del subgraph.blocks[simple_snarl.end]

        return subgraph.blocks, SnarlGraph(
            subgraph.blocks, subgraph.adj_list, simple_snarl.id,
            parent=parent_snarl_graph,
            children=simple_snarl.children,
            start_node=simple_snarl.start,
            end_node=simple_snarl.end
        )

    def length(self):
        if self._length is not None:
            return self._length
        self._length = self._get_longest_path_length()
        return self._length

    def _get_longest_path_length(self):
        memo = self._create_path_length_dict()
        return memo[self._end_node]

    def _create_path_length_dict(self, forward=True):
        if forward:
            start_node = self._start_node
            end_node = self._end_node
            next_node_func = self.get_next_nodes
        else:
            start_node = -self._end_node
            end_node = -self._start_node
            next_node_func = self.get_previous_nodes

        stack = deque([(start_node, 0, 0)])
        memo = {}
        while stack:
            node_id, dist, n = stack.pop()
            next_nodes = next_node_func(node_id)
            for next_node in next_nodes:
                if next_node in memo and memo[next_node] >= dist:
                    continue
                memo[next_node] = dist
                if next_node == end_node:
                    continue
                new_dist = dist + self.node_size(next_node)
                stack.append((next_node, new_dist, n+1))

        return memo

    def get_linear_node_intervals(self):
        linear_node_intervals = {}
        forward_length_dict = self._create_path_length_dict()
        back_length_dict = self._create_path_length_dict(False)
        for node_id in self.blocks:
            if (node_id in forward_length_dict) and (-node_id in back_length_dict):
                start_length = forward_length_dict[node_id]
                end_length = back_length_dict[-node_id]
            elif (-node_id in forward_length_dict) and (node_id in back_length_dict):
                # Use reverse node
                start_length = forward_length_dict[-node_id]
                end_length = back_length_dict[node_id]
            else:
                logging.error(node_id)
                logging.error(forward_length_dict)
                logging.error(back_length_dict)
                raise Exception()

            path_length = start_length + end_length + self.node_size(node_id)
            scale_factor = self.length() / path_length
            if scale_factor < 1:
                raise Exception()

            linear_node_intervals[node_id] = (
                start_length * scale_factor,
                (start_length+self.node_size(node_id))*scale_factor)

        return linear_node_intervals

    def get_distance_dicts(self):
        starts_dict, ends_dict = {}, {}
        linear_node_intervals = self.get_linear_node_intervals()
        for node_id, block in self.blocks.items():
            # TODO: Also get reverse distances
            start, end = linear_node_intervals[node_id]
            if isinstance(block, obg.Block):
                starts_dict[node_id] = start
                ends_dict[node_id] = end
            else:
                sub_starts_dict, sub_ends_dict = block.get_distance_dicts()
                starts_dict.update(
                    {n_id: start+dist
                     for n_id, dist in sub_starts_dict.items()})
                ends_dict.update({n_id: start + dist
                                  for n_id, dist in sub_ends_dict.items()})
        for node_id in starts_dict.keys():
            assert starts_dict[node_id] != ends_dict[node_id]
        return starts_dict, ends_dict

    def find_node_ids(self, offset):
        sub_nodes = [node_id for node_id, interval
                     in self.linear_node_intervals.items()
                     if interval[0] < offset and interval[1] > offset]
        positions = []
        for node_id in sub_nodes:
            if isinstance(self.blocks[node_id],
                          obg.Block):
                positions.append(node_id)
            else:
                positions.extend(self.blocks[node_id].find_node_ids())

        return positions

    def to_file(self, file_name):
        with open("%s" % file_name, "wb") as f:
            pickle.dump(self, f)

    @classmethod
    def from_file(cls, file_name):
        with open("%s" % file_name, "rb") as f:
            obj = pickle.loads(f.read())
            assert isinstance(obj, cls)
            return obj


class SimpleSnarl():
    def __init__(self, start, end, id, parent=None, children=None):
        self.start = start
        self.end = end
        self.id = id
        self.parent = parent

        if children is None:
            self.children = []
        else:
            self.children = children

        assert self.parent != self.id

    def add_child(self, child):
        assert child.parent == self.id
        assert child.id != self.id
        self.children.append(child)

    def set_parent(self, parent_id):
        self.parent = parent_id

    def __str__(self):
        return "Snarl(%d, %d, id=%d, parent=%s)" % (
            self.start, self.end, self.id, self.parent)

    def __repr__(self):
        return self.__str__()

    def sanitize(self):
        for child in self.children:
            assert child.id != self.id, " Child %d has same id as self %d" % (child.id, self.id)


class SnarlGraphBuilder:

    def __init__(self, graph, snarls, id_counter=0):
        assert isinstance(snarls, dict), \
            "Should be dict with ids as keys and SimpleSnarls as values"
        assert id_counter > 0,\
            "ID counter must be set to higher than highest ID in graph"
        self.graph = graph
        self.snarls = snarls  # dict of snarls, id: SimpleSnarls
        self.id_counter = id_counter

    def build_snarl_graphs(self):
        logging.info("Finding top level snarls")
        top_level_snarls = [snarl for snarl in self.snarls.values()
                            if snarl.parent is None]

        # Find all snarls with 0 nodes
        logging.info("Finding snarls with 0 nodes")
        n_zero_nodes = 0
        for snarl in self.snarls.values():
            start = snarl.start
            end = snarl.end

            if end in self.graph.adj_list[start]:
                n_zero_nodes += 1

        # Add dummy start and end to graph
        new_start = self.id_counter
        new_end = new_start + 1

        for block in self.graph.get_first_blocks():
            self.graph._add_edge(new_start, block)

        for block in self.graph.get_last_blocks():
            self.graph._add_edge(block, new_end)

        logging.info("%d snarls with 0 nodes" % n_zero_nodes)
        logging.info("%d top level" % len(top_level_snarls))
        logging.info("%d snarls total" % len(self.snarls))

        whole_graph_snarl = SnarlGraph(self.graph.blocks,
                                       self.graph.adj_list,
                                       "top_level", parent=None,
                                       children=top_level_snarls,
                                       start_node=new_start,
                                       end_node=new_end)
        return whole_graph_snarl

    @classmethod
    def from_vg_snarls(cls, graph, vg_snarls_file_name):
        logging.info("Building snarlgraph from vg graph")
        snarls = Snarls.from_vg_snarls_file(vg_snarls_file_name).snarls
        start_end_mapping = {}
        simple_snarls = {}

        max_graph_id = max([id for id in graph.blocks.keys()])
        id_counter = max_graph_id + 1
        i = 0
        for snarl in snarls:

            assert snarl.start.backward == snarl.end.backward
            start = snarl.start.node_id
            end = snarl.end.node_id

            if snarl.start.backward:
                start = snarl.end.node_id
                end = snarl.start.node_id

            simple_snarls[id_counter] = SimpleSnarl(
                start, end, id=id_counter, parent=snarl.parent)

            start_end_mapping["%d-%d" % (start, end)] = id_counter
            id_counter += 1

            if i % 500000 == 0:
                logging.info("Reading snarl %d from vg" % i)
            i += 1

        logging.info("Setting parent snarls")
        # Set parent
        for id, snarl in simple_snarls.items():
            if snarl.parent is not None:
                parent = snarl.parent
                if parent.start.node_id == 0:
                    snarl.set_parent(None)
                else:
                    mapping_id = "%d-%d" % (
                        parent.start.node_id, parent.end.node_id)
                    if mapping_id in start_end_mapping:
                        parent_id = start_end_mapping[mapping_id]
                        parent = simple_snarls[parent_id]
                        snarl.set_parent(parent.id)
                        assert snarl.parent == parent.id
                        parent.add_child(snarl)
                        assert snarl in parent.children
                        assert parent.id != snarl.id

                    else:
                        logging.warning("Found with parent that is missing")
                        logging.warning(snarl)

        logging.info("Done reading snarls")
        return cls(graph, simple_snarls, id_counter)



import numpy as np


class SubGraphAnalyzer:
    def __init__(self, subgraph, max_path, dists):
        self._subgraph = subgraph
        self._score = np.min(dists[:, -1])
        self._max_path = max_path
        self._dists = dists

    def get_info(self):
        return (self.has_two_bindings(), self.is_ambiguous())

    def get_pruned_graph(self):
        to_dists = self.dists[self._max_path, :]
        from_dists = self.dists[:, -1]
        return np.flatnonzero(to_dists+from_dists == self._score)

    def has_two_bindings(self):
        if len(self._max_path) <= 1:
            return False
        sizes = np.min(self._subgraph, 1)
        return np.sum(sizes) != self._score

    def is_ambiguous(self):
        if len(self._max_path) <= 1:
            return False
        for node in self._max_path[:-1]:
            to_dist = self._dists[self._max_path[0], node]
            row = self._subgraph[node]
            for next_node, d in zip(row.indices, row.data):
                if next_node in self._max_path:
                    continue
                s = to_dist + d + self._dists[next_node, -1]
                assert s >= self._score
                if s == self._score:
                    return True

        return False



import numpy as np


class DensePileup:
    def __init__(self, graph, values):
        self._values = values
        self._graph = graph

    def values_in_range(self, node, start, end):
        index = node - self._graph.min_node
        array_start = self._graph.node_indexes[index] + start
        array_end = self._graph.node_indexes[index] + end
        assert array_end > array_start
        assert array_end <= len(self._values), "Array end %d > len of values %s" % (array_end, len(self._values))
        out = self._values[array_start:array_end]
        assert len(out) > 0
        return out

    def values(self, node_id):
        index = node_id - self._graph.min_node
        array_start = self._graph.node_indexes[index]
        array_end = self._graph.node_indexes[index+1]
        return self._values[array_start:array_end]

    def get_interval_values(self, interval):
        interval_length = interval.length()
        assert interval_length > 0, "Trying to get value of interval with negative length, %s" % interval
        values = np.zeros(interval_length)
        offset = 0

        is_reverse = False
        if interval.region_paths[0] < 0:
            assert np.all(np.array(interval.region_paths) < 0), " First region path negative, but not rest. Interval: %s" % interval
            is_reverse = True
            interval = interval.get_reverse()

        for i, rp in enumerate(interval.region_paths):
            assert rp > 0, "Currently only implemented for forward directed intervals"
            start = 0
            end = self._graph.node_size(rp)
            if i == 0:
                start = interval.start_position.offset
            if i == len(interval.region_paths) - 1:
                end = interval.end_position.offset

            assert end > start, "Start >= end for interval %s" % interval
            values_in_rp = self.values_in_range(rp, start, end)
            values[offset:offset + (end - start)] = values_in_rp

            offset += end-start


        if is_reverse:
            out = values[::-1]
        else:
            out = values

        assert np.sum(np.isnan(values)) == 0, "%s contains nan" % (values)
        return values

    def __str__(self):
        out = "Minimum DensePileup. Values: %s. Sum: %.3f" % (self._values, np.sum(self._values))
        return out

    def __repr__(self):
        return out
